"""Redis kvlayer storage implementation.

.. This software is released under an MIT/X11 open source license.
   Copyright 2014 Diffeo, Inc.

Purpose
=======

This is an implementation of kvlayer that uses redis_ for its
underlying storage.  This generally limits storage to what can be held
in memory on the host system, but at the same time, redis is expected
to be available on many systems and requires much less setup than a
distributed Bigtable system.

Implementation Notes
====================

Redis is also used by rejester_ for its underlying storage.  Using
a different ``app_name`` for the kvlayer storage configuration will avoid
conflicts between the two systems.

A kvlayer "namespace" is a single row named ``APPNAME_NAMESPACE``.
This is a hash mapping kvlayer table names to redis row names.  A
kvlayer "table" is stored in two rows: the mapped table name with no
suffix is a hash mapping serialized UUID tuples to values, and the
mapped table name plus "k" is a sorted set of key names (only, all
with score 0, to support :meth:`RedisStorage.scan`).

This currently uses pure-ASCII key values (using
:func:`kvlayer._utils.join_key_fragments`) and intermediate table row IDs.
Redis should in principle be able to handle packed-binary values,
which would be a quarter the size.

.. _redis: http://redis.io
.. _rejester: https://github.com/diffeo/rejester

Module Contents
===============

"""

from __future__ import absolute_import

import logging
import uuid

import redis

from kvlayer._abstract_storage import AbstractStorage
from kvlayer._exceptions import BadKey, ProgrammerError
from kvlayer._utils import join_key_fragments, split_key, make_start_key, make_end_key

logger = logging.getLogger(__name__)

# Some useful Lua fragments
verify_lua = '''
if redis.call('exists', KEYS[1]) == 0
then return redis.error_reply('no such kvlayer table') end
'''
verify_lua_failed = 'no such kvlayer table'

class RedisStorage(AbstractStorage):
    def __init__(self):
        """Initialize a redis-based storage instance.

        Uses the global kvlayer configuration, with the following parameters:

        .. code-block:: yaml

            namespace: a_namespace
        
        namespace prefix for this storage layer

        .. code-block:: yaml

            app_name: app

        application name prefix for this storage layer

        .. code-block:: yaml

            storage_addresses: ["redis.example.com:6379"]

        list of ``hostname:port`` pairs for redis (only first is used)

        .. code-block:: yaml

            redis_db_num: 1

        Redis database number (defaults to 0)

        """
        super(RedisStorage, self).__init__()
        storage_addresses = self._config.get('storage_addresses', [])
        db_num = self._config.get('redis_db_num', 0)
        if len(storage_addresses) == 0:
            raise ProgrammerError('config lacks storage_addresses')
        if len(storage_addresses) > 1:
            logger.warning('multiple storage_addresses, only first will be used')
        address = storage_addresses[0]
        if ':' in address:
            (host, port) = address.split(':')
            conn_kwargs = { 'host': host, 'port': int(port), 'db': db_num }
        else:
            conn_kwargs = { 'host': address, 'db': db_num }

        logger.debug('will connect to redis {!r}'.format(conn_kwargs))
        self._pool = redis.ConnectionPool(**conn_kwargs)
        self._table_keys = {}
        self._table_sizes = {}
        pass

    def _connection(self):
        """Get a connection to Redis."""
        return redis.StrictRedis(connection_pool=self._pool)

    @property
    def _namespace_key(self):
        """Name of the redis key holding the namespace information."""
        return self._app_name + '_' + self._namespace

    def _table_key(self, conn, table):
        """Get the redis key name of some table.

        If it is in :attr:`_table_keys` then use that value; otherwise
        look it up in redis.  Caches successful lookups.

        :param conn: Redis connection
        :type conn: :class:`redis.StrictRedis`
        :param str table: kvlayer table name
        :return: Name of the redis key holding `table`
        :rtype: str or None

        """
        if table not in self._table_keys:
            k = conn.hget(self._namespace_key, table)
            if k is not None: self._table_keys[table] = k
        return self._table_keys.get(table, None)

    def setup_namespace(self, table_names):
        """Creates tables in the namespace.

        :param table_names: Table names to create
        :type table_names: dictionary mapping string table names to
          int key lengths

        """
        conn = self._connection()
        # We will do this by generating a short random key, and
        # assigning an empty hash to it.  If we can do this successfully
        # then record the mapping.
        #
        # For simplicity we use UUIDs for the keys.  We can almost
        # certainly get away with shorter (or binary) keys.
        script = conn.register_script("""
        local existing = redis.call("hget", KEYS[2], ARGV[1])
        if existing then
          return redis.status_reply(existing)
        elseif redis.call("exists", KEYS[1]) == 0 then
          redis.call("hset", KEYS[1], "", "")
          redis.call("hset", KEYS[2], ARGV[1], KEYS[1])
          return redis.status_reply(KEYS[1])
        else
          return redis.error_reply(KEYS[1] .. " already exists")
        end
        """)
        for table in table_names.keys():
            tries = 5
            while True:
                key = uuid.uuid4().hex
                logger.debug("setup_namespace: table {} trying uuid {} in {}"
                             .format(table, key, self._namespace_key))
                try:
                    key = script(keys=[key, self._namespace_key], args=[table])
                    # will return a key name (possibly an existing key)
                    # or raise ResponseError
                    logger.debug("setup_namespace: table {} uuid {}"
                                 .format(table, key))
                    self._table_keys[table] = key
                    self._table_sizes[table] = table_names[table]
                    break
                except redis.ResponseError, exc:
                    if tries == 0:
                        raise
                    tries -= 1
                    pass # try again with a new uuid
        self.normalize_namespaces(self._table_sizes)
        return

    def delete_namespace(self):
        """Deletes all data from the namespace."""
        conn = self._connection()
        # To encourage atomicity, we will do this by
        # 1a. Get the table rows from the master row
        # 1b. Delete the master row
        # 2. Delete the invidual table rows
        script = conn.register_script("""
        local vals = redis.call('hvals', KEYS[1])
        redis.call('del', KEYS[1])
        return vals
        """)
        table_keys = script(keys=[self._namespace_key])
        if len(table_keys) > 0:
            conn.delete(*table_keys)
            conn.delete(*[k + 'k' for k in table_keys])
        self._table_keys = {}
        self._table_sizes = {}

    def clear_table(self, table_name):
        """Delete all data from one table.

        :param str table_name: Name of the kvlayer table
        :raise kvlayer._exceptions.BadKey: `table_name` does not exist
          in this namespace

        """
        conn = self._connection()
        key = self._table_key(conn, table_name)
        if key is None:
            raise BadKey(table_name)
        logger.debug('clear_table %s in namespace %s', table_name,
                     self._namespace_key)
        # There is a potential race condition the the namespace
        # is deleted *right here*.  We trap that (if the namespace
        # is deleted then the individual rows will be deleted).
        # So if you get the BadKey later on, that's why.
        script = conn.register_script(verify_lua + '''
        redis.call('del', KEYS[1])
        redis.call('del', KEYS[2])
        redis.call('hset', KEYS[1], '', '')
        ''')
        try:
            script(keys=[key, key + 'k'])
        except redis.ResponseError, exc:
            if str(exc) == verify_lua_failed:
                raise BadKey(table_name)
            raise

    def put(self, table_name, *keys_and_values, **kwargs):
        """Save values for keys in `table_name`.

        Each key must be a tuple of UUIDs for the length specified
        for `table_name` in :meth:`setup_namespace`.

        :param str table_name: Name of the kvlayer table
        :param keys_and_values: Data to add
        :type keys_and_values: pairs of (key tuple, value)
        :param int batch_size: accepted but ignored
        :raise kvlayer._exceptions.BadKey: `table_name` does not exist
          in this namespace

        """
        conn = self._connection()
        key = self._table_key(conn, table_name)
        if key is None:
            raise BadKey(table_name)
        params = []
        for (k,v) in keys_and_values:
            #logger.debug('put {} {!r} {}'.format(table_name, k, v))
            key_spec = self._table_sizes[table_name]
            ex = self.check_put_key_value(k, v, table_name,
                                          key_spec)
            if ex is not None:
                raise ex
            params.append(join_key_fragments(k, key_spec=key_spec))
            params.append(v)
        script = conn.register_script(verify_lua + '''
        for i = 1, #ARGV, 2 do
          redis.call('hset', KEYS[1], ARGV[i], ARGV[i+1])
          redis.call('zadd', KEYS[2], 0, ARGV[i])
        end
        return redis.status_reply(KEYS[1])
        ''')
        try:
            script(keys=[key, key + 'k'], args=params)
        except redis.ResponseError, exc:
            if str(exc) == verify_lua_failed:
                raise BadKey(table_name)
            raise

    def scan(self, table_name, *key_ranges, **kwargs):
        """Yield pairs from selected ranges in some table.

        `key_ranges` are pairs of tuples of :class:`uuid.UUID`,
        specifying valid lower and upper bounds for uuid-tuple keys.
        All keys in the table that are at least the lower bound of
        some range and at most the upper bound of the same range are
        returned.  If no `key_ranges` are provided, scan the entire
        table.

        >>> for (k,v) in storage.scan('table',
        ...                           ((start1,None),(end1,None)),
        ...                           ((start2,None),(end2,None))):
        ...   print k

        :param str table_name: Name of the kvlayer table
        :param key_ranges: Ranges to scan
        :type key_ranges: pairs of uuid tuples
        :raise kvlayer._exceptions.BadKey: `table_name` does not exist
          in this namespace

        """
        conn = self._connection()
        key = self._table_key(conn, table_name)
        if key is None:
            raise BadKey(table_name)
        #logger.debug('scan {} {!r}'.format(table_name, key_ranges))
        key_spec = self._table_sizes[table_name]
        if len(key_ranges) == 0:
            # scan the whole table
            # just do this in one big call for simplicity
            res = conn.hgetall(key)
            for k in sorted(res.iterkeys()):
                if k == '': continue
                uuids = split_key(k, key_spec)
                yield (uuids, res[k])
        for start, end in key_ranges:
            find_first = '''
            local first = redis.call('zrank', KEYS[2], ARGV[1])
            if not first then
              redis.call('zadd', KEYS[2], 0, ARGV[1])
              first = redis.call('zrank', KEYS[2], ARGV[1])
              redis.call('zrem', KEYS[2], ARGV[1])
            end
            '''
            no_first = '''
            local first = 0
            '''
            find_last = '''
            local last = redis.call('zrank', KEYS[2], ARGV[2])
            if not last then
              redis.call('zadd', KEYS[2], 0, ARGV[2])
              last = redis.call('zrank', KEYS[2], ARGV[2])
              redis.call('zrem', KEYS[2], ARGV[2])
              if last == 0 then return {} end
              last = last - 1
            end
            '''
            no_last = '''
            local last = -1
            '''
            do_scan = '''
            local keys = redis.call('zrange', KEYS[2], first, last)
            local result = {}
            for i = 1, #keys do
              result[i*2-1] = keys[i]
              result[i*2] = redis.call('hget', KEYS[1], keys[i])
            end
            return result
            '''
            script = verify_lua
            if start:
                script += find_first
            else:
                script += no_first
            if end:
                script += find_last
            else:
                script += no_last
            script += do_scan
            script = conn.register_script(script)
            try:
                res = script(keys=[key, key+'k'],
                             args=[make_start_key(start, key_spec=key_spec),
                                   make_end_key(end, key_spec=key_spec)])
            except redis.ResponseError, exc:
                if str(exc) == verify_lua_failed:
                    raise BadKey(table_name)
                raise
            keys = res[0::2]
            values = res[1::2]
            for k,v in zip(keys, values):
                uuids = split_key(k, key_spec)
                yield (uuids, v)

    def get(self, table_name, *keys, **kwargs):
        """Yield specific pairs from the table.

        `keys` are tuples of :class:`uuid.UUID`.  Yields pairs
        of (uuids,value) for any present values.

        >>> for (k,v) in storage.get('table', key):
        ...   assert k == key
        ...   print v

        :param str table_name: Name of the kvlayer table
        :param keys: Keys to fetch
        :type key_ranges: tuples of :class:`uuid.UUID`
        :raise kvlayer._exceptions.BadKey: `table_name` does not exist
          in this namespace

        """
        # We can be sufficiently atomic without lua scripting here.
        # The hmget call is atomic, so if the table gets deleted or
        # cleared in between checking for the table's existence
        # (which, remember, could be cached) and actually fetching keys,
        # then this yields nothing, which is consistent.
        if not keys:
            raise StopIteration
        conn = self._connection()
        key = self._table_key(conn, table_name)
        if key is None:
            raise BadKey(key)
        key_spec = self._table_sizes[table_name]
        #logger.debug('get {} {!r}'.format(table_name, keys))
        ks = [join_key_fragments(k, key_spec=key_spec) for k in keys]
        vs = conn.hmget(key, *ks)
        for (k, v) in zip(ks, vs):
            # v may be None if the key isn't there; yield it anyways
            yield (tuple(split_key(k, key_spec)),v)
            
    def delete(self, table_name, *keys, **kwargs):
        """Delete specific pairs from the table.

        `keys` are pairs of (uuids,value); the uuids are tuples of
        :class:`uuid.UUID`, and the values are ignored.

        >>> storage.delete('table', (key,None))

        :param str table_name: Name of the kvlayer table
        :param keys: Keys to delete
        :type key_ranges: pairs of (uuids,any)
        :raise kvlayer._exceptions.BadKey: `table_name` does not exist
          in this namespace

        """
        # Again blow off atomicity.  The worst that happens is that
        # the entire table is deleted in between getting its name and
        # deleting single rows from it, and in that case since the
        # entire table is deleted, all of the specific things we wanted
        # gone, are gone.
        if not keys:
            return
        conn = self._connection()
        key = self._table_key(conn, table_name)
        if key is None:
            raise BadKey(table_name)
        logger.debug('delete {} {!r}'.format(table_name, keys))
        key_spec = self._table_sizes[table_name]
        ks = [join_key_fragments(k, key_spec=key_spec) for k in keys]
        script = conn.register_script(verify_lua + '''
        for i = 1, #ARGV do
          redis.call('hdel', KEYS[1], ARGV[i])
          redis.call('zrem', KEYS[2], ARGV[i])
        end
        ''')
        try:
            script(keys=[key, key+'k'], args=ks)
        except redis.ResponseError, exc:
            if str(exc) == verify_lua_failed:
                raise BadKey(table_name)
            raise

    def close(self):
        """Close connections and end use of this storage client."""
        self._pool.disconnect()



'''Database abstraction for key/value stores.

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.

Many popular large-scale databases export a simple key/value
abstraction: the database is simply a list of cells with some
(possibly structured) key and a value for each key.  This allows the
database system itself to partition the database in some way, and in a
distributed system it allows the correct system hosting a specific key
to be found easily.  This model is also simple enough that it can be
used with in-memory storage or more traditional SQL-based databases.

This module provides a simple abstraction around these
key/value-oriented databases.  It works with :mod:`yakonfig` to hold
basic configuration settings, so the top-level YAML configuration must
have a ``kvlayer`` block.  This will typically look something like

.. code-block:: yaml

    kvlayer:
      storage_type: redis
      storage_addresses: [redis.example.com:6379]
      app_name: app
      namespace: namespace

These four parameters are always required.  ``storage_type`` gives one
of the database backends described below.  ``storage_addresses`` is a
list of backend-specific database locations.  ``app_name`` and
``namespace`` combine to form a container for the virtual tables
stored in kvlayer.

Backends
========

local
-----

This is intended only for testing.  Values are stored in a Python
dictionary and not persisted.  ``storage_addresses`` are not required.

.. code-block:: yaml

    kvlayer:
      storage_type: local

filestorage
-----------

This is intended only for testing.  Values are stored in a local file
using the :mod:`shelve` module.  This does not require
``storage_addresses``, but it does require:

.. code-block:: yaml

    kvlayer:
      storage_type: filestorage

      # Name of the file to use for storage
      filename: /tmp/kvlayer.bin

      # If set, actually work on a copy of "filename" at this location.
      copy_to_filename: /tmp/kvlayer-copy.bin

redis
-----

Uses the `Redis`_ in-memory database.

.. code-block:: yaml

    kvlayer:
      storage_type: redis

      # host:port locations of Redis servers; only the first is used
      storage_addresses: [redis.example.com:6379]

      # Redis database number (default: 0)
      redis_db_num: 1

accumulo
--------

Uses the Apache `Accumulo`_ distributed database.  Your installation
must be running the Accumulo proxy, and the configuration points at
that proxy.

.. code-block:: yaml

    kvlayer:
      storage_type: accumulo

      # host:port location of the proxy; only the first is used
      storage_addresses: [accumulo.example.com:50096]
      username: root
      password: secret

      # all of the following parameters are default values and are optional
      accumulo_max_memory: 1000000
      accumulo_timeout_ms: 30000
      accumulo_threads: 10
      accumulo_latency_ms: 10
      thrift_framed_transport_size_in_mb: 15

Each kvlayer table is instantiated as an Accumulo table named
``appname_namespace_table``.

.. _Accumulo: http://accumulo.apache.org/

postgres
--------

Uses `PostgreSQL`_ for storage.  This backend is only available if the
:mod:`psycopg2` module is installed.  The ``storage_type`` is a
`PostgreSQL connection string`_.  The ``app_name`` and ``namespace``
can only consist of alphanumeric characters, underscores, or ``$``,
and must begin with a letter or underscore.

.. code-block:: yaml

    kvlayer:
      storage_type: postgres
      storage_addresses:
      - 'host=postgres.example.com port=5432 user=test dbname=test password=test'

The backend assumes the user is able to run SQL ``CREATE TABLE`` and
``DROP TABLE`` statements.  Each kvlayer namespace is instantiated as
an SQL table named ``kv_appname_namespace``; kvlayer tables are
collections of rows within the namespace table sharing a common field.

.. _PostgreSQL: http://www.postgresql.org
.. _PostgreSQL connection string: http://www.postgresql.org/docs/current/static/libpq-connect.html#LIBPQ-PARAMKEYWORDS

cassandra
---------

Uses the Apache `Cassandra`_ distributed database.

.. code-block:: yaml

    kvlayer:
      storage_type: cassandra
      storage_addresses: ['cassandra.example.com:9160']
      username: root
      password: secret

      connection_pool_size: 2
      max_consistency_delay: 120
      replication_factor: 1
      thrift_framed_transport_size_in_mb: 15

.. _Cassandra: http://cassandra.apache.org/

API
===

Having set up the global configuration, it is enough to call
:func:`kvlayer.client` to get a storage client object.

The API works in terms of "tables", though these are slightly
different from tradational database tables.  Each table has keys which
are tuples of a fixed length.  Tuple members must be
:class:`uuid.UUID` objects unless the system is configured with
``keys_must_be_uuid: false``.

.. autofunction:: client

.. autoclass:: kvlayer._abstract_storage.AbstractStorage
   :members:
   :undoc-members:
   :show-inheritance:

.. autoclass:: DatabaseEmpty
.. autoclass:: BadKey

Instance Collections
====================

.. automodule:: kvlayer.instance_collection

'''

from kvlayer._client import client
from kvlayer.config import config_name, default_config, add_arguments, \
    runtime_keys, check_config
from kvlayer._exceptions import DatabaseEmpty, BadKey



"""test _util key functions

Your use of this software is governed by your license agreement.

Copyright 2012-2014 Diffeo, Inc.

"""

import pytest

from kvlayer._exceptions import BadKey
from kvlayer._utils import join_key_fragments

def test_join_key_fragments():
    assert join_key_fragments(('a', 'b')) == 'a\x00b'

    assert join_key_fragments(('a', 2), key_spec=(str,int)) == 'a\x0000000000000000000000000000000002'
    assert join_key_fragments(('a', 2), key_spec=(str,(int,long))) == 'a\x0000000000000000000000000000000002'
    assert join_key_fragments(('a', 0x12341234123412341234), key_spec=(str,long)) == 'a\x0000000000000012341234123412341234'
    assert join_key_fragments(('a', 0x12341234123412341234), key_spec=(str,(int,long))) == 'a\x0000000000000012341234123412341234'

    with pytest.raises(BadKey):
        join_key_fragments(('a', 'b'), key_spec=(str,int))
    with pytest.raises(BadKey):
        join_key_fragments(('a', 0x12341234123412341234), key_spec=(str,int))

    assert join_key_fragments(('a foo', 2), key_spec=(str,int)) == 'a foo\x0000000000000000000000000000000002'
    with pytest.raises(BadKey):
        join_key_fragments(('a\x00foo', 2), key_spec=(str,int))




'''
Implementation of AbstractStorage using local memory
instead of a DB backend

Your use of this software is governed by your license agreement.

Copyright 2012-2014 Diffeo, Inc.
'''

import abc
import logging
import time

from kvlayer._abstract_storage import AbstractStorage
from kvlayer._utils import _requires_connection, make_start_key, make_end_key, join_key_fragments

logger = logging.getLogger(__name__)

class AbstractLocalStorage(AbstractStorage):
    """Local in-memory storage for testing.

    This is a base class for storage implementations that use a
    dictionary-like object for actually holding data.

    """

    def __init__(self):
        super(AbstractLocalStorage, self).__init__()
        self._connected = False
        self._raise_on_missing = self._config.get('raise_on_missing', True)

    def setup_namespace(self, table_names):
        '''creates tables in the namespace.  Can be run multiple times with
        different table_names in order to expand the set of tables in
        the namespace.
        '''
        logger.debug('creating tables: %r', table_names)
        self._table_names.update(table_names)
        self.normalize_namespaces(self._table_names)
        ## just store everything in a dict
        for table in table_names:
            if table not in self._data:
                self._data[table] = dict()
        self._connected = True

    @_requires_connection
    def clear_table(self, table_name):
        self._data[table_name] = dict()

    @_requires_connection
    def put(self, table_name, *keys_and_values, **kwargs):
        start_time = time.time()
        keys_size = 0
        values_size = 0
        num_keys = 0
        for key, val in keys_and_values:
            ex = self.check_put_key_value(key, val, table_name, self._table_names[table_name])
            if ex:
                raise ex
            self._data[table_name][key] = val
            if self._log_stats is not None:
                num_keys += 1
                keys_size += len(join_key_fragments(key))
                values_size += len(val)

        end_time = time.time()
        num_values = num_keys

        self.log_put(table_name, start_time, end_time, num_keys, keys_size, num_values, values_size)

    @_requires_connection
    def scan(self, table_name, *key_ranges, **kwargs):
        start_time = time.time()
        num_keys = 0
        keys_size = 0
        values_size = 0

        key_spec = self._table_names[table_name]
        key_ranges = list(key_ranges)
        specific_key_range = True
        if not key_ranges:
            key_ranges = [[None, None]]
            specific_key_range = False
        for start, finish in key_ranges:
            total_count = 0
            start = make_start_key(start, key_spec=key_spec)
            finish = make_end_key(finish, key_spec=key_spec)
            for key in sorted(self._data[table_name].iterkeys()):
                ## given a range, mimic the behavior of DBs that tell
                ## you if they failed to find a key
                ## LocalStorage does get/put on the Python tuple as the key, stringify for sort comparison
                joined_key = join_key_fragments(key, key_spec=key_spec)
                if (start is not None) and (start > joined_key):
                    continue
                if (finish is not None) and (finish < joined_key):
                    continue
                total_count += 1
                val = self._data[table_name][key]
                yield key, val

                if self._log_stats is not None:
                    keys_size += len(join_key_fragments(key))
                    values_size += len(val)
                    num_keys += 1
                    values_size += len(val)

        end_time = time.time()
        num_values = num_keys
        self.log_scan(table_name, start_time, end_time, num_keys, keys_size, num_values, values_size)

    @_requires_connection
    def get(self, table_name, *keys, **kwargs):
        start_time = time.time()
        num_keys = 0
        keys_size = 0
        num_values = 0
        values_size = 0

        for key in keys:
            if self._log_stats is not None:
                num_keys += 1
                keys_size += len(join_key_fragments(key))
            try:
                key, value = key, self._data[table_name][key]
                yield key, value
                num_values += 1
                values_size += len(value)
            except KeyError:
                yield key, None

        end_time = time.time()
        self.log_get(table_name, start_time, end_time, num_keys, keys_size, num_values, values_size)

    @_requires_connection
    def delete(self, table_name, *keys):
        start_time = time.time()
        num_keys = 0
        keys_size = 0

        for key in keys:
            if self._log_stats is not None:
                num_keys += 1
                keys_size += len(join_key_fragments(key))
            self._data[table_name].pop(key, None)

        end_time = time.time()
        self.log_delete(table_name, start_time, end_time, num_keys, keys_size)

    def close(self):
        ## prevent reading until connected again
        self._connected = False

class LocalStorage(AbstractLocalStorage):
    """Local in-memory storage for testing.

    All instances of LocalStorage share the same underlying data.
    """

    _data = {}

    def delete_namespace(self):
        self._data.clear()



import logging
import os
import re
import sys
import uuid

from pyaccumulo import Accumulo, Mutation
import pytest
import yaml

import kvlayer
from kvlayer._accumulo import AStorage, _string_decrement
import kvlayer.tests.make_namespace
import yakonfig

logger = logging.getLogger(__name__)

@pytest.fixture(scope='module')
def config_path(request):
    return str(request.fspath.dirpath('config_accumulo.yaml'))

@pytest.yield_fixture
def direct(config_path, namespace_string):
    with yakonfig.defaulted_config([kvlayer], filename=config_path,
                                   params={'app_name': 'kvlayer',
                                           'namespace': namespace_string}):
        config = yakonfig.get_global_config('kvlayer')
        conn = Accumulo(host='test-accumulo-1.diffeo.com', port=50096,
                        user=config['username'], password=config['password'])

        yield conn

        tables = conn.list_tables()
        for table in tables:
            if re.search(namespace_string, table):
                conn.delete_table(table)

@pytest.yield_fixture(scope='function', params=['accumulo'])
def client(namespace_string, config_path):
    app_name = 'kvlayer'
    with yakonfig.defaulted_config([kvlayer], filename=config_path,
                                   params={'app_name': app_name,
                                           'namespace': namespace_string}):
        logger.info('initializing client')
        client = kvlayer.client()
        def _test_ns(name):
            return '_'.join([app_name, namespace_string, name])
        client._test_ns = _test_ns

        yield client

        logger.info('tearing down %s', _test_ns(''))
        client.delete_namespace()
        logger.info('done cleaning up')

def test_setup_namespace(client, direct):
    #storage = AStorage()
    logger.info('creating namespace: %r' % client._namespace)
    client.setup_namespace({'table1': 1, 'table2': 1})
    logger.info('checking existence of tables')
    assert direct.table_exists(client._test_ns('table1'))
    assert direct.table_exists(client._test_ns('table2'))
    logger.info('finished checking')


def test_delete_namespace(client, direct):
    storage = AStorage()
    client.setup_namespace({'table1': 1, 'table2': 1})
    tables = direct.list_tables()
    assert client._test_ns('table1') in tables
    assert client._test_ns('table2') in tables
    storage.delete_namespace()
    tables = direct.list_tables()
    assert client._test_ns('table1') not in tables
    assert client._test_ns('table2') not in tables


def test_clear_table(client, direct):
    client.setup_namespace({'table1': 1, 'table2': 1})

    # Write some rows to table
    m = Mutation('row_1')
    m.put(cf='cf1', cq='cq1', val='1')
    m.put(cf='cf1', cq='cq1', val='2')
    direct.write(client._test_ns('table1'), m)

    # Clear table
    client.clear_table('table1')

    # Verify clear
    for entry in direct.scan(client._test_ns('table1')):
        assert False

    # Clear an empty table
    client.clear_table('table1')

    # Verify still clear
    for entry in direct.scan(client._test_ns('table1')):
        assert False


def test_put_get(client, direct):
    client.setup_namespace({'table1': 1, 'table2': 1})
    kv_dict = {(uuid.uuid4(),): 'value' + str(x) for x in xrange(10)}
    keys_and_values = [(key, value) for key, value in kv_dict.iteritems()]
    client.put('table1', *keys_and_values)
    keys = kv_dict.keys()
    values = kv_dict.values()
    for entry in direct.scan(client._test_ns('table1')):
        logger.debug('found %r:%r', entry.row, entry.val)
        assert entry.val in values
    generator = client.scan('table1', (keys[0], keys[0]))
    key, value = generator.next()
    assert kv_dict[key] == value
    for x in xrange(10):
        for key, value in client.scan('table1', (keys[x], keys[x])):
            assert kv_dict[key] == value


def test_scan_all_keys(client, direct):
    client.setup_namespace({'table1': 1, 'table2': 1})
    kv_dict = {(uuid.uuid4(),): 'value' + str(x) for x in xrange(10)}
    keys_and_values = [(key, value) for key, value in kv_dict.iteritems()]
    client.put('table1', *keys_and_values)

    for key, val in client.scan('table1'):
        assert kv_dict[key] == val
        del kv_dict[key]

    assert kv_dict == {}


def test_delete(client, direct):
    client.setup_namespace({'table1': 1})
    kv_dict = {(uuid.uuid4(),): 'value' + str(x) for x in xrange(10)}
    keys_and_values = [(key, value) for key, value in kv_dict.iteritems()]
    client.put('table1', *keys_and_values)
    keys = kv_dict.keys()
    delete_keys = keys[0:len(keys):2]
    save_keys = keys[1:len(keys):2]
    assert set(delete_keys)
    client.delete('table1', *delete_keys)
    for key in save_keys:
        for key, value in client.scan('table1', (key, key)):
            assert kv_dict[key] == value
    for key in delete_keys:
        assert list(client.scan('table1', (key, key))) == []


def test_close(client):
    conn = client.conn
    assert conn
    assert client._connected
    client.close()
    assert not client._connected


def test_string_decrement():
    assert _string_decrement('b') == 'a\xff'
    assert _string_decrement('b\0') == 'a\xff\xff'
    assert _string_decrement('\0') == None
    assert _string_decrement('') == None
    assert _string_decrement('b\xff') == 'b\xfe\xff'



'''
Definition of AbstractStorage, which all storage
implementations inherit.

Your use of this software is governed by your license agreement.

Copyright 2012-2014 Diffeo, Inc.
'''

from __future__ import absolute_import
import abc
import atexit
import itertools
import operator
import time
import uuid

from kvlayer._exceptions import BadKey, ProgrammerError
import yakonfig

class AbstractStorage(object):
    '''Base class for all low-level storage implementations.

    All of the table-like structures we use are setup like this::

        namespace = dict(
            table_name = dict((UUID, UUID, ...): val)
            ...
        )

    where the number of UUIDs in the key is a configurable parameter
    of each table, and the "val" is always binary and might be a
    trivial value, like 1.

    '''
    __metaclass__ = abc.ABCMeta

    def check_put_key_value(self, key, value, table_name, key_spec):
        "check that (key, value) are ok. return Exception or None if okay."
        if not isinstance(key, tuple):
            return BadKey('key should be tuple, but got %s' % (type(key),))
        if len(key) != len(key_spec):
            return BadKey('%r wants %r parts in key tuple, but got %r' % (table_name,  len(key_spec), len(key)))
        for kp, ks in zip(key, key_spec):
            if not isinstance(kp, ks):
                return BadKey('part of key wanted type %s but got %s: %r' % (ks, type(kp), kp))
        return None

    @abc.abstractmethod
    def __init__(self):
        '''Initialize a storage instance with config dict.
        Typical config fields:
        'namespace': string name of set of tables this kvlayer instance refers to
        'app_name': string name of application code which is connecting
        'storage_addresses': [list of server specs]
        'username'
        'password'
        '''
        self._config = yakonfig.get_global_config('kvlayer')
        self._table_names = {}
        self._namespace = self._config.get('namespace', None)
        if not self._namespace:
            raise ProgrammerError('kvlayer requires a namespace')
        self._app_name = self._config.get('app_name', None)
        if not self._app_name:
            raise ProgrammerError('kvlayer requires an app_name')
        self._require_uuid = self._config.get('keys_must_be_uuid', True)
        log_stats_cfg = self._config.get('log_stats', None)
        # StorageStats also consumes:
        #  log_stats_interval_ops
        #  log_stats_interval_seconds
        if log_stats_cfg:
            self._log_stats = StorageStats(log_stats_cfg, self._config)
        else:
            self._log_stats = None

    @abc.abstractmethod
    def setup_namespace(self, table_names):
        '''Create tables in the namespace.

        Can be run multiple times with different `table_names` in
        order to expand the set of tables in the namespace.  This
        generally needs to be called by every client, even if only
        reading data.

        Tables are specified by the form of their keys. A key must be
        a tuple of a set number and type of parts. Currently types
        :class:`uuid.UUID`, :class:`int`, :class:`long`, and
        :class:`str` are well supported, anything else is serialzed by
        :func:`str`. Historically, a kvlayer key had to be a tuple of some
        number of UUIDs.  `table_names` is a dictionary mapping a
        table name to a tuple of types.  The dictionary values may also
        be integers, in which case the tuple is that many UUIDs.

        :param dict table_names: Mapping from table name to value type tuple

        '''
        return

    def normalize_namespaces(self, table_names):
        '''Normalize table_names spec dictionary in place.

        Replaces ints with a tuple of that many (uuid.UUID,)
        '''
        for k, v in table_names.iteritems():
            if isinstance(v, (int, long)):
                assert v < 50, "assuming attempt at very long key is a bug"
                table_names[k] = (uuid.UUID,) * v

    @abc.abstractmethod
    def delete_namespace(self):
        '''Deletes all data from namespace.'''
        return

    @abc.abstractmethod
    def clear_table(self, table_name):
        '''Delete all data from one table.'''
        return

    @abc.abstractmethod
    def put(self, table_name, *keys_and_values, **kwargs):
        '''Save values for keys in `table_name`.

        Each key must be a tuple of length and types as specified for
        `table_name` in :meth:`setup_namespace`.

        '''
        return

    def log_put(self, table_name, start_time, end_time, num_keys, keys_size, num_values, values_size):
        if self._log_stats is not None:
            self._log_stats.put.add(table_name, start_time, end_time, num_keys, keys_size, num_values, values_size)

    @abc.abstractmethod
    def scan(self, table_name, *key_ranges, **kwargs):
        '''Yield tuples of (key, value) from querying table_name for
        items with keys within the specified ranges.  If no key_ranges
        are provided, then yield all (key, value) pairs in table.
        This may return nothing if the table is empty or there are
        no matching keys in any of the specified ranges.

        Each of the `key_ranges` is a pair of a start and end tuple to
        scan.  To specify the beginning or end, a -Inf or Inf value,
        use an empty tuple as the beginning or ending key of a range.

        '''
        return

    def log_scan(self, table_name, start_time, end_time, num_keys, keys_size, num_values, values_size):
        if self._log_stats is not None:
            self._log_stats.scan.add(table_name, start_time, end_time, num_keys, keys_size, num_values, values_size)

    def scan_keys(self, table_name, *key_ranges, **kwargs):
        '''Scan only the keys from a table.

        Yields key tuples from queying `table_name` for keys within
        the specified ranges.  If no `key_ranges` are provided, then
        yield all key tuples in the table.  This may yield nothing if
        the table is empty or there are no matching keys in any of the
        specified ranges.

        Each of the `key_ranges` is a pair of a start and end tuple to
        scan.  To specify the beginning or end, a -Inf or Inf value,
        use an empty tuple as the beginning or ending key of a range.

        '''
        # Feel free to reimplement this if your backend can do better!
        # we don't do log_scan_keys() here, let underlying scan() call log_scan()
        return itertools.imap(operator.itemgetter(0),
                              self.scan(table_name, *key_ranges, **kwargs))

    def log_scan_keys(self, table_name, start_time, end_time, num_keys, keys_size):
        if self._log_stats is not None:
            self._log_stats.scan_keys.add(table_name, start_time, end_time, num_keys, keys_size, 0, 0)

    @abc.abstractmethod
    def get(self, table_name, *keys, **kwargs):

        '''Yield tuples of (key, value) from querying table_name for items
        with keys.  If any of the key tuples are not in the table,
        those key tuples will be yielded with value :const:`None`.

        '''
        return

    def log_get(self, table_name, start_time, end_time, num_keys, keys_size, num_values, values_size):
        if self._log_stats is not None:
            self._log_stats.get.add(table_name, start_time, end_time, num_keys, keys_size, num_values, values_size)

    @abc.abstractmethod
    def delete(self, table_name, *keys, **kwargs):
        '''Delete all (key, value) pairs with specififed keys

        '''
        return

    def log_delete(self, table_name, start_time, end_time, num_keys, keys_size):
        if self._log_stats is not None:
            self._log_stats.delete.add(table_name, start_time, end_time, num_keys, keys_size, 0, 0)

    @abc.abstractmethod
    def close(self):
        '''
        close connections and end use of this storage client
        '''
        if self._log_stats is not None:
            self._log_stats.close()
            self._log_stats = None


# I may have built this inside-out.
# Maybe the top level split should be on table, and keep stats of ops within that.
# It shouldn't be hard to transpose for display if needed.
class StorageStats(object):
    def __init__(self, config_str_or_writeable, config):
        self._f = None
        self._config_str = None
        if hasattr(config_str_or_writeable, 'write'):
            self._f = config_str_or_writeable
        elif isinstance(config_str_or_writeable, (str, unicode)):
            self._config_str = config_str_or_writeable
        self._interval_ops = config.get('log_stats_interval_ops')
        if self._interval_ops is not None:
            self._interval_ops = int(self._interval_ops)
            if self._interval_ops <= 0:
                self._interval_ops = None
        self._op_interval_counter = 0
        self._interval_seconds = config.get('log_stats_interval_seconds')
        if self._interval_seconds is not None:
            self._interval_seconds = float(self._interval_seconds)
            if self._interval_seconds <= 0.0:
                self._interval_seconds = None
        self._interval_s_last_flushed = time.time()
        self.put = OpStats(self)
        self.scan = OpStats(self)
        self.scan_keys = OpStats(self)
        self.get = OpStats(self)
        self.delete = OpStats(self)

        self._closed = False
        atexit.register(self.atexit)

    def __str__(self):
        return 'put:\n{x.put}\nscan:\n{x.scan}\nscan_keys:\n{x.scan_keys}\nget:\n{x.get}\ndelete:\n{x.delete}\n'.format(x=self)

    def _out(self):
        if (self._f is None) and hasattr(self._config_str, 'write'):
            self._f = self._config_str
        if self._f is None:
            self._f = open(self._config_str, 'a')
        return self._f

    def did_op(self):
        if self._closed:
            return
        if self._interval_ops is not None:
            self._op_interval_counter += 1
            if self._op_interval_counter >= self._interval_ops:
                self.flush()
                self._interval_s_last_flushed = time.time()
                self._op_interval_counter = 0
                return
        if self._interval_seconds is not None:
            now = time.time()
            if self._interval_s_last_flushed + self._interval_seconds < now():
                self.flush()
                self._interval_s_last_flushed = now
                self._op_interval_counter = 0
                return

    def flush(self):
        out = self._out()
        out.write(time.strftime('%Y%m%d_%H%M%S\n'))
        out.write(self.__str__())
        if hasattr(out, 'flush'):
            out.flush()

    def close(self):
        if self._f is not None:
            if hasattr(self._f, 'close'):
                self._f.close()
            self._f = None
        self._closed = True

    def atexit(self):
        if not self._closed:
            self.flush()
            self.close()


class OpStats(object):
    def __init__(self, parentStorageStats):
        self.by_table = {}
        self.num_ops = 0
        self._psto = parentStorageStats

    def add(self, table_name, start_time, end_time, num_keys, keys_size, num_values, values_size):
        self.num_ops += 1
        ts = self.by_table.get(table_name)
        if ts is None:
            ts = OpStatsPerTable()
            self.by_table[table_name] = ts
        ts.add(start_time, end_time, num_keys, keys_size, num_values, values_size)
        self._psto.did_op()

    def __str__(self):
        parts = []
        total = OpStatsPerTable()
        for k,v in self.by_table.iteritems():
            total += v
            parts.append('{:10s} {}\n'.format(k, str(v)))
        return ''.join(parts) + '           {}\n'.format(str(total))


class OpStatsPerTable(object):
    def __init__(self):
        self.total_time = 0.0
        self.num_ops = 0
        self.num_keys = 0
        self.keys_size = 0
        self.num_values = 0
        self.values_size = 0

    def add(self, start_time, end_time, num_keys, keys_size, num_values, values_size=0):
        self.total_time += (end_time - start_time)
        self.num_ops += 1
        self.num_keys += num_keys
        self.keys_size += keys_size
        self.num_values += num_values
        self.values_size += values_size

    def __iadd__(self, b):
        assert isinstance(b, OpStatsPerTable)
        self.total_time += b.total_time
        self.num_ops += b.num_ops
        self.num_keys += b.num_keys
        self.keys_size += b.keys_size
        self.num_values += b.num_values
        self.values_size += b.values_size
        return self

    def __str__(self):
        out = '{t:0.3f}s on {ops} ops ({spops:0.6g} s/op), {k} keys for {kb} bytes ({bpk} B/key)'.format(
            t=self.total_time,
            ops=self.num_ops,
            spops=self.num_ops and (self.total_time/self.num_ops),
            k=self.num_keys,
            kb=self.keys_size,
            bpk=self.num_keys and ((1.0*self.keys_size)/self.num_keys)
        )
        if self.num_values > 0:
            out += '{v} values for {vb} bytes ({bpv} B/val)'.format(
                v=self.num_values,
                vb=self.values_size,
                bpv=(1.0*self.values_size)/self.num_values
            )
        return out



#!/usr/bin/env python
"""kvlayer self-tests.

-----

Your use of this software is governed by your license agreement.

Copyright 2014 Diffeo, Inc.

"""
from __future__ import absolute_import
import argparse
import os
import sys

import pytest

def main():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('redis_address', metavar='HOST:PORT',
                        help='location of a redis instance to use for testing')
    args = parser.parse_args()
    test_dir = os.path.dirname(__file__)
    response = pytest.main(['-v', '-v',
                            '--redis-address', args.redis_address,
                            '-k', 'not (cassandra or accumulo or postgres)',
                            test_dir])
    sys.exit(response)

if __name__ == '__main__':
    main()



import pytest
import logging

logger = logging.getLogger(__name__)

config_postgres = {
    'namespace': 'test',
    'storage_addresses': ['host=test-postgres.diffeo.com port=5432 user=test dbname=test password=test'],
}


try:
    from kvlayer._postgres import PGStorage, _valid_namespace, detatch_on_exception
    postgres_missing = 'False'
except ImportError:
    postgres_missing = 'True'

@pytest.mark.skipif(postgres_missing)
@pytest.mark.parametrize(
    "badnamespace",
    [None,
     '9isnotaletter',
     '$isnotaletter',
 ])
def test_illegal_namespaces(badnamespace):
    with pytest.raises(Exception):
        config = dict(config_postgres)
        config['namespace'] = badnamespace
        pg = PGStorage(config)


@pytest.mark.skipif(postgres_missing)
@pytest.mark.parametrize(
    "namespace",
    ['_ok', 'Aok', 'aok'])
def test_legal_namespaces(namespace):
    assert _valid_namespace(namespace)



class Closeable(object):
    def __init__(self):
        self.closed = False

    @detatch_on_exception
    def err(self):
        raise Exception('err')

    def close(self):
        self.closed = True


@pytest.mark.skipif(postgres_missing)
def test_detatch_on_exception():
    a = Closeable()
    with pytest.raises(Exception):
        a.err()
    assert a.closed


class NotCloseable(object):
    @detatch_on_exception
    def errX(self):
        raise Exception('err')

@pytest.mark.skipif(postgres_missing)
def test_detatch_on_exception2():
    # check that @detatch_on_exception doesn't panic when there's no self.close
    a = NotCloseable()
    with pytest.raises(Exception):
        a.errX()



from __future__ import division, absolute_import
import argparse
import multiprocessing
import logging
import Queue
from signal import alarm, signal, SIGHUP, SIGTERM, SIGABRT, SIGALRM
import sys
import uuid
import time

import pytest

import kvlayer
from kvlayer.tests.test_interface import client, backend # fixture
import yakonfig

logger = logging.getLogger(__name__)

def worker(operation, i_queue, o_queue, tasks_remaining,
           class_config=None,
    ):
    '''simple worker used by run_many below

@operation(task, o_queue): callable that takes "task" received from
i_queue and optionally passes something to o_queue

@i_queue: a multiprocessing Queue provided by run_many
@o_queue: a multiprocessing Queue provided by run_many
@tasks_remaining: a multiprocessing Event provided by run_many

@class_config: if not None, then operation is initialized by calling
operation(class_config) before using it inside each worker.

    '''
    if class_config is not None:
        ## operation is a class with __call__ so initialize it
        operation = operation(class_config)
    while 1:
        try:
            task = i_queue.get(timeout=0.3)
            operation(task, o_queue)
        except Queue.Empty:
            if tasks_remaining.is_set():
                continue
            else:
                break

def run_many(operation, task_generator, num_workers=1, timeout=None,
             class_config=None,
    ):
    '''test harness that runs num_workers as child processes and sends
them tasks from task_generator via queue.  Results are harvested from
o_queue and yielded by this function, so typical usage is:

for response from run_many(do_search, queries, num_workers=10, timeout=5):
    validate(response)

@operation: defined above

@task_generator: an iterable of things to pass through i_queue to workers

@num_workers: number of child processes to run in parallel, default is 1

@timeout: (required) tests must set an explicit timeout in seconds

@class_config: defined above
    '''
    if timeout is None:
        raise Exception('tests must set an explicit timeout')

    task_generator = iter(task_generator)

    manager = multiprocessing.Manager()
    tasks_remaining = manager.Event()
    tasks_remaining.set()

    ## make I/O queues with no size constraint
    i_queue = manager.Queue() # pylint: disable=E1101
    o_queue = manager.Queue() # pylint: disable=E1101

    pool = multiprocessing.Pool(num_workers, maxtasksperchild=1)

    ## prepare to catch control-C
    global run_many_not_killed
    run_many_not_killed = True
    def clean_exit(sig_num, frame):
        global run_many_not_killed
        run_many_not_killed = False
        logger.critical('attempting clean_exit')
        pool.terminate()
        logger.critical('pool terminated')
        pool.join()
        logger.critical('pool joined. Exiting.')
    for sig_num in [SIGTERM, SIGHUP, SIGABRT, SIGALRM]:
        signal(sig_num, clean_exit)

    ## load up the pool
    async_results = []
    for x in range(num_workers):
        async_res = pool.apply_async(
            worker, 
            args=(operation, i_queue, o_queue, tasks_remaining, class_config))
        async_results.append(async_res)

    start_time = time.time()
    while run_many_not_killed and async_results:
        elapsed = time.time() - start_time
        if elapsed > timeout:
            raise Exception('%.1f seconds elapsed' % elapsed)
        for async_res in async_results:
            try:
                async_res.get(0)
            except multiprocessing.TimeoutError:
                logger.info('worker in progress')
            except Exception, exc:
                logger.info('worker died!', exc_info=True)
                raise
            else:
                logger.info('worker finished')
                assert async_res.ready()
                async_results.remove(async_res)
        while elapsed < timeout:
            elapsed = time.time() - start_time
            try:
                resp = o_queue.get(block=False)
                yield resp
            except Queue.Empty:
                # logger.info('no responses from o_queue')
                time.sleep(0.2)
                break
        tasks_added = 0
        while o_queue.qsize() < num_workers * 2 and elapsed < timeout \
              and tasks_added < num_workers * 2:
            elapsed = time.time() - start_time
            try:
                task = task_generator.next()
                i_queue.put(task)
                tasks_added += 1
            except StopIteration:
                tasks_remaining.clear()
                break

    pool.close()
    pool.join()
    logger.info('finished running %d worker processes' % num_workers)


def pass_through(task, o_queue):
    o_queue.put(task)

def sleeper(task, o_queue):
    time.sleep(4)

def test_multiprocessing_harness():
    num_workers = 10
    task_generator = range(100)
    ret_vals = list(run_many(pass_through, task_generator, 
                             num_workers=num_workers, timeout=20))
    assert set(ret_vals) == set(range(100)), set(range(100)) - set(ret_vals)

def test_multiprocessing_harness_timeout():
    num_workers = 10
    task_generator = range(100)
    with pytest.raises(Exception) as exc:
        list(run_many(sleeper, task_generator, 
                      num_workers=num_workers, timeout=2))
    assert 'seconds elapsed' in str(exc)

def test_multiprocessing_harness_control_C():
    num_workers = 10
    task_generator = range(100)
    alarm(1)
    ret_vals = list(run_many(sleeper, task_generator, 
                             num_workers=num_workers, timeout=20))
    assert len(ret_vals) == 0


class random_inserts(object):
    def __init__(self, config):
        self.client = kvlayer.client()
        self.client.setup_namespace(dict(t1=1))
        self.one_mb = ' ' * 2**20

    def __call__(self, u, o_queue):
        try:
            self.client.put('t1', ((u,), self.one_mb))
        except Exception, exc:
            logger.critical('client failed!', exc_info=True)
            raise exc
        o_queue.put(u)
        logger.info('put one_mb at %r', u)


class many_gets(object):
    def __init__(self, config):
        self.client = kvlayer.client()
        self.client.setup_namespace(dict(t1=1))

    def __call__(self, u, o_queue):
        try:
            kvs = list(self.client.get('t1', (u,)))
        except Exception, exc:
            logger.critical('client failed!', exc_info=True)
            raise exc
        assert len(kvs[0][1]) == 2**20
        o_queue.put((True, u))
        logger.info('retrievied one_mb at %r', u)


def test_throughput_insert_random(client, num_workers=5, num_inserts=100):
    client.setup_namespace(dict(t1=1))
    
    total_inserts = num_workers * num_inserts
    task_generator = [uuid.uuid4() for x in xrange(total_inserts)]
    start_time = time.time()
    ret_vals = list(run_many(random_inserts, task_generator, 
                             class_config=client._config,
                             num_workers=num_workers, timeout=total_inserts * 5))
    elapsed = time.time() - start_time
    assert len(ret_vals) == total_inserts
    rate = total_inserts / elapsed
    logger.info('%d MB written in %.1f seconds --> %.1f MB/sec across %d workers for storage_type=%s',
                total_inserts, elapsed, rate, num_workers, client._config['storage_type'])

    if client._config['storage_type'] in ['postgres', 'accumulo', 'cassandra', 'redis']:
        start_time = time.time()
        count = 0
        for (found, u) in run_many(many_gets, ret_vals, 
                                   class_config=client._config,
                                   num_workers=num_workers, timeout=total_inserts * 5):
            if not found:
                raise Exception('failed to find %r' % u)
            count += 1
        elapsed = time.time() - start_time
        assert len(ret_vals) == total_inserts
        rate = total_inserts / elapsed
        logger.info('%d MB read in %.1f seconds --> %.1f MB/sec across %d workers for storage_type=%s',
                    total_inserts, elapsed, rate, num_workers, client._config['storage_type'])


class indexer(object):
    '''
    rapidly changes an index in t2
    '''
    def __init__(self, config):
        self.client = kvlayer.client()
        self.client.setup_namespace(dict(t1=1, t2=2))

    def __call__(self, u, o_queue):
        try:
            kvs = list(self.client.scan('t2', ((u,), (u,))))
        except Exception, exc:
            logger.critical('client failed!', exc_info=True)
            raise exc
        u1, u2 = kvs[0][0]
        assert u == u1

        #... need to populate t2 initially, and then flow records through it from 0 to 1-8, to 9-24, 25-57, 58-121, etc.


class joiner(object):
    def __init__(self, config):
        self.client = kvlayer.client()
        self.client.setup_namespace(dict(t1=1, t2=2))

    def __call__(self, u, o_queue):
        try:
            kvs = list(self.client.scan('t2', ((u,), (u,))))
        except Exception, exc:
            logger.critical('client failed!', exc_info=True)
            raise exc
        u1, u2 = kvs[0][0]
        assert u == u1
        try:
            kvs = list(self.client.get('t1', (u2,)))
        except Exception, exc:
            logger.critical('client failed!', exc_info=True)
            raise exc
        assert len(kvs[0][1]) == 2**20
        o_queue.put((True, u2))
        logger.info('retrievied one_mb at %r', u2)

@pytest.mark.skipif('True')
def test_throughput_join(client):
    '''measure throughput of reading data from t1 by first looking up the
    key in t2, while t2 is under write load
    '''
    client.setup_namespace(dict(t1=1, t2=2))
    
    num_workers = 10
    num_inserts = 100
    total_inserts = num_workers * num_inserts
    data_ids = [uuid.uuid4() for x in xrange(total_inserts)]
    start_time = time.time()
    ret_vals = list(run_many(random_inserts, data_ids, 
                             class_config=client._config,
                             num_workers=num_workers, timeout=total_inserts/2))
    elapsed = time.time() - start_time
    assert len(ret_vals) == total_inserts
    rate = total_inserts / elapsed
    logger.info('%d MB written in %.1f seconds --> %.1f MB/sec using %d parallel workers for storage_type=%s',
                total_inserts, elapsed, rate, num_workers, client._config['storage_type'])

    ## start tool subprocesses that each run a pool
    #writers = multiprocessing.Process(
        #target=run_many,
        #args=(indexer, data_ids, 
              #class_config=client._config,
              #num_workers=num_workers, timeout=total_inserts/2))))

    ### finish writing this test...

def main():
    parser = argparse.ArgumentParser(usage='kvlayer_throughput_test <storage_type> <storage_address1> <storage_address2> ...',
                                     conflict_handler='resolve')
    parser.add_argument('storage_type')
    parser.add_argument('storage_addresses', default=[], action='append')
    parser.add_argument('--app-name', default='kvlayer')
    parser.add_argument('--namespace', default='test_throughput')
    parser.add_argument('--num-workers', default=5, type=int)
    parser.add_argument('--num-inserts', default=100, type=int)
    modules = [yakonfig, kvlayer]
    args = yakonfig.parse_args(parser, modules)
    #config = yakonfig.get_global_config()

    logging.basicConfig(level=logging.DEBUG)

    if not args.storage_addresses:
        sys.exit('must specify at least one storage_address')

    with yakonfig.defaulted_config([kvlayer], params=vars(args)):
        client = kvlayer.client()
        client.delete_namespace()
        test_throughput_insert_random(
            client, 
            num_workers=args.num_workers, num_inserts=args.num_inserts,
            )
        client.delete_namespace()






class KVLayerError(Exception):
    pass


class ConfigurationError(KVLayerError):
    pass


class ProgrammerError(KVLayerError):
    pass


class StorageClosed(KVLayerError):
    pass


class DatabaseEmpty(KVLayerError):
    pass


class BadKey(KVLayerError):
    """A key value passed to a kvlayer function was not of the correct form.

    Keys must be tuples of a fixed length of :class:`uuid.UUID` objects.
    The length of the tuple is specified in the initial call to
    :meth:`kvlayer._abstract_storage.AbstractStorage.setup_namespace`.

    """
    pass

class SerializationError(KVLayerError):
    """Converting between an item and a serialized form failed."""
    pass



from __future__ import absolute_import
import itertools
import logging
from operator import attrgetter
import os
import getpass
import random
import sys
import time
import uuid

import cql
import pycassa
from pycassa.pool import ConnectionPool
from pycassa.system_manager import SystemManager, SIMPLE_STRATEGY, \
    LEXICAL_UUID_TYPE, ASCII_TYPE, BYTES_TYPE
from pycassa.types import CompositeType, TimeUUIDType, LexicalUUIDType, UUIDType, UTF8Type
import pytest
import yaml

import kvlayer
import yakonfig

logger = logging.getLogger(__name__)
one_mb = ' ' * 2**20

@pytest.yield_fixture(scope='function')
def client(namespace_string, request):
    config_path = str(request.fspath.dirpath('config_cassandra.yaml'))
    with yakonfig.defaulted_config([kvlayer], filename=config_path,
                                   params={'namespace': namespace_string,
                                           'app_name': 'kvltest'}):
        client = kvlayer.client()
        yield client
        client.delete_namespace()

def grouper(iterable, n, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx
    args = [iter(iterable)] * n
    return itertools.izip_longest(fillvalue=fillvalue, *args)

@pytest.mark.skipif('True')
def test_composite_column_names(client):
    '''
    examine the unique nature of cassandras "wide sorted rows";

    Basically, C* is good at storing large OrderedDict objects and
    less good at the simpler kind of key=value storage that we were
    doing earlier.
    '''
    config = client._config
    namespace = client._app_namespace
    chosen_server = client._chosen_server
    sm = SystemManager(chosen_server)
    sm.create_keyspace(namespace, SIMPLE_STRATEGY, {'replication_factor': '1'})

    family = 'test'
    sm.create_column_family(
        namespace, family, super=False,
        key_validation_class = ASCII_TYPE,
        default_validation_class = BYTES_TYPE,
        comparator_type=CompositeType(UUIDType(), UUIDType()),
        #column_name_class = LEXICAL_UUID_TYPE
        )

    logger.info( sm.describe_schema_versions() )

    pool = ConnectionPool(namespace, config['storage_addresses'],
                          max_retries=1000, pool_timeout=10, pool_size=2, timeout=120)

    cf = pycassa.ColumnFamily(pool, family)
    u1, u2, u3, u4 = uuid.uuid1(), uuid.uuid1(), uuid.uuid1(), uuid.uuid1()

    ## insert four
    cf.insert('inbound', {(u1, u2): b''})
    cf.insert('inbound', {(u1, u3): b''})
    cf.insert('inbound', {(u1, u4): b''})
    cf.insert('inbound', {(u1, u1): b'45'})

    ## insert three with duplicates
    cf.insert('inbound', {(u3, u3): b'42'})
    cf.insert('inbound', {(u3, u3): b'43'})
    cf.insert('inbound', {(u3, u2): b''})
    cf.insert('inbound', {(u3, u4): b''})

    ## insert two
    cf.insert('inbound', {(u2, u3): b''})
    cf.insert('inbound', {(u2, u4): b''})

    ## verify start/finish parameters work as expected
    assert 4 == len(cf.get('inbound', column_start=(u1,), column_finish=(u1,)))
    assert 2 == len(cf.get('inbound', column_start=(u2,), column_finish=(u2,)))
    assert 3 == len(cf.get('inbound', column_start=(u3,), column_finish=(u3,)))

    ## check delete of a single column
    cf.remove('inbound', columns=[(u2, u3)])
    assert 1 == len(cf.get('inbound', column_start=(u2,), column_finish=(u2,)))

    ## check that the last inserted value appears at the expected
    ## location in the OrderedDict
    assert '43' == cf.get('inbound', column_start=(u3,), column_finish=(u3,)).items()[1][1]

    rec1 = cf.get('inbound', column_start=(u3,u3), column_finish=(u3,u3)).items()
    assert len(rec1) == 1

    start  = uuid.UUID(int=0)
    finish = uuid.UUID(int=2**128-1)
    assert start < u1 < u2 < u3 < u4 < finish
    assert 8 == len(cf.get('inbound', column_start=(start,), column_finish=(finish,)).items())

    ## test range searching
    start  = uuid.UUID(int=u3.int - 1)
    finish = uuid.UUID(int=u3.int + 1)
    assert start.int < u3.int < finish.int
    rec2 = cf.get('inbound', column_start=(start,), column_finish=(finish,)).items()
    assert rec2[1][0] == rec1[0][0]
    assert rec2[1][1] == rec1[0][1] == '43'

    cf.insert('inbound', {(u3,u3): b''.join(map(lambda u: u.bytes, (u1,u2,u3,u4)))})
    data = cf.get('inbound', column_start=(u3,u3), column_finish=(u3,u3)).items()[0][1]

    assert [u1,u2,u3,u4] == map(lambda b: uuid.UUID(bytes=b), grouper(data, 16))

    sm.close()

@pytest.mark.skipif('True')
def test_composite_column_names_with_decomposited_keys(client):
    '''
    examine the unique nature of cassandras "wide sorted rows" using
    concatenated keys
    '''
    config = client._config
    namespace = client._app_namespace
    chosen_server = client._chosen_server
    sm = SystemManager(chosen_server)
    sm.create_keyspace(namespace, SIMPLE_STRATEGY, {'replication_factor': '1'})

    family = 'test'
    sm.create_column_family(
        namespace, family, super=False,
        key_validation_class = ASCII_TYPE,
        default_validation_class = BYTES_TYPE,
        comparator_type=UTF8Type()
        )

    logger.info( sm.describe_schema_versions() )

    pool = ConnectionPool(namespace, config['storage_addresses'],
                          max_retries=1000, pool_timeout=10, pool_size=2, timeout=120)

    cf = pycassa.ColumnFamily(pool, family)
    u1, u2, u3, u4 = uuid.uuid1(), uuid.uuid1(), uuid.uuid1(), uuid.uuid1()

    ## insert four
    cf.insert('inbound', {join_uuids(u1, u2): b''})
    cf.insert('inbound', {join_uuids(u1, u3): b''})
    cf.insert('inbound', {join_uuids(u1, u4): b''})
    cf.insert('inbound', {join_uuids(u1, u1): b'45'})

    ## insert three with duplicates
    cf.insert('inbound', {join_uuids(u3, u3): b'42'})
    cf.insert('inbound', {join_uuids(u3, u3): b'43'})
    cf.insert('inbound', {join_uuids(u3, u2): b''})
    cf.insert('inbound', {join_uuids(u3, u4): b''})

    ## insert two
    cf.insert('inbound', {join_uuids(u2, u3): b''})
    cf.insert('inbound', {join_uuids(u2, u4): b''})

    ## verify start/finish parameters work as expected
    assert 4 == len(cf.get('inbound', column_start=join_uuids(u1, num_uuids=2), column_finish=join_uuids(u1, num_uuids=2, padding='f')))
    assert 2 == len(cf.get('inbound', column_start=join_uuids(u2, num_uuids=2), column_finish=join_uuids(u2, num_uuids=2, padding='f')))
    assert 3 == len(cf.get('inbound', column_start=join_uuids(u3, num_uuids=2), column_finish=join_uuids(u3, num_uuids=2, padding='f')))

    ## check delete of a single column
    cf.remove('inbound', columns=[join_uuids(u2, u3)])
    assert 1 == len(cf.get('inbound', column_start=join_uuids(u2, num_uuids=2), column_finish=join_uuids(u2, num_uuids=2, padding='f')))

    ## check that the last inserted value appears at the expected
    ## location in the OrderedDict
    assert '43' == cf.get('inbound', column_start=join_uuids(u3, num_uuids=2), column_finish=join_uuids(u3, num_uuids=2, padding='f')).items()[1][1]

    rec1 = cf.get('inbound', column_start=join_uuids(u3,u3), column_finish=join_uuids(u3,u3)).items()
    assert len(rec1) == 1

    start  = uuid.UUID(int=0)
    finish = uuid.UUID(int=2**128-1)
    assert start < u1 < u2 < u3 < u4 < finish
    assert 8 == len(cf.get('inbound', column_start=join_uuids(start, num_uuids=2), column_finish=join_uuids(finish, num_uuids=2, padding='f')).items())

    ## test range searching
    start  = uuid.UUID(int=u3.int - 1)
    finish = uuid.UUID(int=u3.int + 1)
    assert start.int < u3.int < finish.int
    rec2 = cf.get('inbound', column_start=join_uuids(start, num_uuids=2), column_finish=join_uuids(finish, num_uuids=2, padding='f')).items()
    assert rec2[1][0] == rec1[0][0]
    assert rec2[1][1] == rec1[0][1] == '43'

    cf.insert('inbound', {join_uuids(u3,u3): b''.join(map(lambda u: u.bytes, (u1,u2,u3,u4)))})
    data = cf.get('inbound', column_start=join_uuids(u3,u3), column_finish=join_uuids(u3,u3)).items()[0][1]

    assert [u1,u2,u3,u4] == map(lambda b: uuid.UUID(bytes=b), grouper(data, 16))

    sm.close()

@pytest.mark.xfail
def test_composite_column_names_second_level_range_query(client):
    '''
    check that we can execute range queries on the second part of a
    CompositeType column name
    '''
    config = client._config
    namespace = client._app_namespace
    chosen_server = client._chosen_server
    sm = SystemManager(chosen_server)
    sm.create_keyspace(namespace, SIMPLE_STRATEGY, {'replication_factor': '1'})

    family = 'test'
    sm.create_column_family(
        namespace, family, super=False,
        key_validation_class = ASCII_TYPE,
        default_validation_class = BYTES_TYPE,
        comparator_type=CompositeType(UUIDType(), UUIDType()),
        )
    pool = ConnectionPool(namespace, config['storage_addresses'],
                          max_retries=1000, pool_timeout=10, pool_size=2, timeout=120)

    cf = pycassa.ColumnFamily(pool, family)
    u1, u2, u3, u4 = uuid.uuid1(), uuid.uuid1(), uuid.uuid1(), uuid.uuid1()

    cf.insert('inbound', {(u1, u2): b''})
    cf.insert('inbound', {(u1, u3): b''})
    cf.insert('inbound', {(u1, u4): b''})

    ## test range searching
    start  = uuid.UUID(int=u3.int - 1)
    finish = uuid.UUID(int=u3.int + 1)
    assert start.int < u3.int < finish.int
    rec3 = cf.get('inbound',
                  column_start =(u1, start),
                  column_finish=(u1, finish)).items()
    assert len(rec3) == 1
    assert rec3[0][0][1] == u3
    ####  This assert above passes!

    ####  This next part fails :-/
    ## now insert many rows -- enough that some should fall in each
    ## subrange below
    for i in xrange(1000):
        cf.insert('inbound', {(u1, uuid.uuid4()): b''})

    ## do four ranges, and expect more than zero in each
    step_size = 2**(128 - 2)
    for i in range(2**2, 0, -1):
        start =  uuid.UUID(int=(i-1) * step_size)
        finish = uuid.UUID(int=min(i * step_size, 2**128 - 1))
        recs = cf.get('inbound',
                      column_start =(u1, start),
                      column_finish=(u1, finish)).items()
        for key, val in recs:
            assert val == b''
            assert key[0] == u1
            assert key[1] < finish
            assert start < key[1]              ## this fails!!

        assert len(recs) > 0
        logger.info( '%r for %r %r' % (len(recs), start, finish))

    sm.close()

# TODO: delete this and use kvlayer._util.join_uuids(), but really drop it and use join_key_fragments
def join_uuids(*uuids, **kwargs):
    num_uuids = kwargs.pop('num_uuids', 0)
    padding = kwargs.pop('padding', '0')
    uuid_str = ''.join(map(attrgetter('hex'), uuids))
    uuid_str += padding * ((num_uuids * 32) - len(uuid_str))
    return uuid_str

# TODO: delete this and use kvlayer._util.split_uuids(), but really use split_key_fragments
def split_uuids(uuid_str):
    return map(lambda s: uuid.UUID(hex=''.join(s)), grouper(uuid_str, 32))

@pytest.mark.skipif('True')
def test_composite_column_names_second_level_range_query_with_decomposited_keys(client):
    '''
    check that we can execute range queries on the second part of a
    CompositeType column name after we unpack the composite key into a
    long string of concatenated hex forms of the UUIDs
    '''
    config = client._config
    namespace = client._app_namespace
    chosen_server = client._chosen_server
    sm = SystemManager(chosen_server)
    sm.create_keyspace(namespace, SIMPLE_STRATEGY, {'replication_factor': '1'})

    family = 'test'
    sm.create_column_family(
        namespace, family, super=False,
        key_validation_class = ASCII_TYPE,
        default_validation_class = BYTES_TYPE,
        comparator_type=UTF8Type(),
        )
    pool = ConnectionPool(namespace, config['storage_addresses'],
                          max_retries=1000, pool_timeout=10, pool_size=2, timeout=120)

    cf = pycassa.ColumnFamily(pool, family)
    u1, u2, u3, u4 = uuid.uuid1(), uuid.uuid1(), uuid.uuid1(), uuid.uuid1()

    cf.insert('inbound', {join_uuids(u1, u2): b''})
    cf.insert('inbound', {join_uuids(u1, u3): b''})
    cf.insert('inbound', {join_uuids(u1, u4): b''})

    ## test range searching
    start  = uuid.UUID(int=u3.int - 1)
    finish = uuid.UUID(int=u3.int + 1)
    assert start.int < u3.int < finish.int
    rec3 = cf.get('inbound',
                  column_start =join_uuids(u1, start),
                  column_finish=join_uuids(u1, finish)).items()
    assert len(rec3) == 1
    assert split_uuids(rec3[0][0])[1] == u3
    ####  This assert above passes!

    ####  This next part fails :-/
    ## now insert many rows -- enough that some should fall in each
    ## subrange below
    for i in xrange(1000):
        cf.insert('inbound', {join_uuids(u1, uuid.uuid4()): b''})

    ## do four ranges, and expect more than zero in each
    step_size = 2**(128 - 2)
    for i in range(2**2, 0, -1):
        start =  uuid.UUID(int=(i-1) * step_size)
        finish = uuid.UUID(int=min(i * step_size, 2**128 - 1))
        recs = cf.get('inbound',
                      column_start =join_uuids(u1, start),
                      column_finish=join_uuids(u1, finish)).items()
        for key, val in recs:
            key = split_uuids(key)
            assert val == b''
            assert key[0] == u1
            assert key[1] < finish
            assert start < key[1]   ## this passes!!

        assert len(recs) > 0
        logger.info( '%r for %r %r' % (len(recs), start, finish))

    sm.close()

@pytest.mark.skipif('True')
def test_cql_paging(client):
    '''
    read rows from a range of tokens in Cassandra
    '''
    config = client._config
    server = config['storage_addresses'][0]
    server = server.split(':')[0]
    conn = cql.connect(server, cql_version='3.0.1')
    cursor = conn.cursor()

    try:
        cursor.execute('DROP KEYSPACE %s;' % client._app_namespace)
    except:
        pass

    ## make a keyspace and a simple table
    ## cursor.execute("CREATE KEYSPACE %s WITH strategy_class = 'SimpleStrategy' AND strategy_options:replication_factor = 1;" % client._app_namespace)
    cursor.execute("CREATE KEYSPACE %s WITH REPLICATION = {'class' : 'SimpleStrategy', 'replication_factor' : 1};" % client._app_namespace)
    cursor.execute("USE %s;" % client._app_namespace)
    cursor.execute('CREATE TABLE data (k int PRIMARY KEY, v varchar);')

    ## put some data in the table
    count = 0
    start_time = time.time()
    for idx in xrange(10):
        cursor.execute("INSERT INTO data (k, v) VALUES (%d, '%s');" % (idx, one_mb))
        count += 1
        if count % 10 == 0:
            elapsed = time.time() - start_time
            rate = float(count) / elapsed
            logger.info( '%d in %.1f sec --> %.1f row/sec' % (count, elapsed, rate) )

    cql3_command = "SELECT k, v FROM data limit 30000;"
    logger.info( cql3_command )

    cursor.execute(cql3_command)

    start = time.time()
    for row in cursor:
        logger.info( time.time() - start )
        start = time.time()

    #print "Number of results: ", len(list(cursor))

    ## remove this test keyspace
    try:
        cursor.execute('DROP KEYSPACE %s;' % client._app_namespace)
    except:
        pass

    cursor.close()
    conn.close()


@pytest.mark.skipif('True')
def test_cql_token_range(client):
    '''
    read rows from a range of tokens in Cassandra
    '''
    config = client._config
    namespace = client._app_namespace
    chosen_server = client._chosen_server
    server = config['storage_addresses'][0]
    server = server.split(':')[0]
    conn = cql.connect(server, cql_version='3.0.1')
    cursor = conn.cursor()

    ## make a keyspace and a simple table
    ## cursor.execute("CREATE KEYSPACE %s WITH strategy_class = 'SimpleStrategy' AND strategy_options:replication_factor = 1;" % client._app_namespace)
    try:
        ## remove this test keyspace
        cursor.execute('DROP KEYSPACE %s;' % client._app_namespace)
    except:
        pass
    cursor.execute("CREATE KEYSPACE %s WITH REPLICATION = {'class' : 'SimpleStrategy', 'replication_factor' : 1};" % client._app_namespace)
    cursor.execute("USE %s;" % client._app_namespace)
    cursor.execute('CREATE TABLE data (k int PRIMARY KEY, v varchar);')

    ## put some data in the table
    count = 0
    total_inbound = 100
    start_time = time.time()
    for idx in xrange(total_inbound):
        cursor.execute("INSERT INTO data (k, v) VALUES (%d, '%s');" % (idx, ' ')) #one_mb))
        count += 1
        if count % 10 == 0:
            elapsed = time.time() - start_time
            rate = float(count) / elapsed
            logger.info( '%d in %.1f sec --> %.1f row/sec' % (count, elapsed, rate) )

    assert count == total_inbound

    ## split up the full range of tokens.
    ## Suppose there are 2**k workers:
    k = 2 # --> four workers
    token_sub_range = 2**(64 - k)
    lowest_token = -1 * 2**63  ## for murmur3hash

    total_count = 0

    for worker_i in range(2**k):
        start_token = lowest_token +      worker_i  * token_sub_range
        end_token =   lowest_token + (1 + worker_i) * token_sub_range
        end_token = min(2**63 - 1, end_token)
        if worker_i == 2**k:
            ## catch the end
            pass#end_token += 1

        limit = 1000
        cql3_command = "SELECT k, v FROM data WHERE token(k) >= %d AND token(k) < %d LIMIT %d;" % (start_token, end_token, limit)
        logger.info( cql3_command )

        cursor.execute(cql3_command)

        start = time.time()
        for row in cursor:
            logger.info( time.time() - start )
            start = time.time()

            total_count += 1

    assert total_count == total_inbound

    ## remove this test keyspace
    cursor.execute('DROP KEYSPACE %s;' % client._app_namespace)

    cursor.close()
    conn.close()



'''
Tools for storage implementations.

Your use of this software is governed by your license agreement.

Copyright 2012-2013 Diffeo, Inc.
'''

import uuid
import itertools
from operator import attrgetter
from kvlayer._exceptions import StorageClosed, BadKey, SerializationError

def _requires_connection(func):
    '''
    Decorator for methods on any implementation of AbstractStorage.
    Raises StorageClosed when self._connected is not True.
    '''
    def wrapped_func(self, *args, **kwargs):
        if not self._connected:
            raise StorageClosed()
        return func(self, *args, **kwargs)
    return wrapped_func


# Now more generally, split a key string into components.
# Non-UUID keys are joined on '\0', so split on that if present.
def split_uuids(uuid_str):
    if '\0' in uuid_str:
        return uuid_str.split('\0')
    return map(lambda s: uuid.UUID(hex=''.join(s)), grouper(uuid_str, 32))


def join_uuids(*uuids):
    '''
    constructs a string by concatenating the hex values of the uuids.
    '''
    if not uuids or uuids[0] == '':
        uuid_str = b''
    else:
        uuid_str = ''.join(map(lambda x: x.hex, uuids))
    return uuid_str


_default_deserializer_map = {
    int : lambda x: int(x, 16),
    long : lambda x: long(x, 16),
    (int, long) : lambda x: long(x, 16),
    uuid.UUID : lambda x: uuid.UUID(hex=x),
    str : lambda x: x,
}


def split_key(key_str, key_spec, splitter='\0'):
    parts = key_str.split(splitter)
    if len(parts) != len(key_spec):
        raise SerializationError('tried to split key into {0} parts but got {1}, from {2!r}'.format(len(key_spec), len(parts), key_str))
    oparts = []
    for i in xrange(len(key_spec)):
        part = parts[i]
        kt = key_spec[i]
        des = _default_deserializer_map.get(kt)
        if des is None:
            raise SerializationError("don't know how to deserialize key part type {0}".format(kt))
        oparts.append(des(part))
    return tuple(oparts)


def _default_uuid_formatter(x):
    return x.hex

def _default_int_formatter(x):
    return '{0:032x}'.format(x)

def default_key_serializer(x):
    if isinstance(x, uuid.UUID) or hasattr(x, 'hex'):
        return _default_uuid_formatter(x)
    if isinstance(x, (int, long)):
        # format an int to the same number of nybbles as a UUID
        return _default_int_formatter(x)
    return str(x)


def _check_types(key_fragments, key_spec, splitter):
    if key_spec:
        for i in xrange(len(key_fragments)):
            kf = key_fragments[i]
            ks = key_spec[i]
            if not isinstance(kf, ks):
                raise BadKey('key[%s] is %s but wanted %s' % (i, type(kf), ks))


def _serialize_check_join(key_fragments, serializer, splitter):
    parts = []
    for x in key_fragments:
        sx = serializer(x)
        if splitter in sx:
            raise BadKey('serialized key fragment %r must not contain special string %r. If this happens frequently, pick a better splitter or implement escaping on serialized key fragments.' % (sx, splitter))
        parts.append(sx)
    return splitter.join(parts)


def join_key_fragments(key_fragments, splitter='\0', key_spec=None, key_serializer=None):
    _check_types(key_fragments, key_spec, splitter)
    return _serialize_check_join(key_fragments, key_serializer or default_key_serializer, splitter)


def make_start_key(key_fragments, key_spec=None, splitter='\0', key_serializer=None):
    '''
    create a byte string key which will be the start of a scan range
    '''
    if key_fragments is None:
        return None
    pfx = join_key_fragments(key_fragments, splitter, key_spec, key_serializer)
    if ((key_spec and
         len(key_fragments) > 0 and
         len(key_fragments) < len(key_spec))):
        pfx += splitter
    print 'make_start_key %r %r %r'%(key_fragments, key_spec, pfx)
    return pfx


def make_uuid_start_key(key_fragments, num_uuids=0):
    parts = [x.hex for x in key_fragments]
    while len(parts) < num_uuids:
        parts.append('00000000000000000000000000000000')
    return ''.join(parts)


def make_end_key(key_fragments, key_spec=None, splitter='\0', key_serializer=None):
    '''
    create a byte string key which will be the end of a scan range
    '''
    if key_fragments is None:
        return None
    pfx = join_key_fragments(key_fragments, splitter, key_spec, key_serializer)
    if key_spec:
        if len(key_fragments) == 0:
            pfx += '\xff'
        if len(key_fragments) < len(key_spec):
            pfx += splitter[:-1] + chr(ord(splitter[-1]) + 1)
        else:
            pfx += '\0'
    else:
        pfx += '\xff'
    print 'make_end_key %r %r %r'%(key_fragments, key_spec, pfx)
    return pfx


def make_uuid_end_key(key_fragments, num_uuids=0):
    parts = [x.hex for x in key_fragments]
    while len(parts) < num_uuids:
        parts.append('ffffffffffffffffffffffffffffffff')
    return ''.join(parts)


def grouper(iterable, n, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx
    args = [iter(iterable)] * n
    return itertools.izip_longest(fillvalue=fillvalue, *args)

class batches(object):
    '''
    return lists of length n drawn from an iterable.  The last batch
    may be shorter than n.
    '''
    def __init__(self, iterable, n):
        self.items = iter(iterable)
        self.n = n

    def __iter__(self):
        batch = []
        for i in self.items:
            batch.append(i)
            if len(batch) == self.n:
                yield batch
                batch = []
        if batch:
            yield batch




import time
import pytest
from kvlayer._decorators import retry

def mocksleep(time):
    return


def test_retry_fails(monkeypatch):

    monkeypatch.setattr(time, 'sleep', mocksleep)

    @retry()
    def raise_forever():
        raise

    with pytest.raises(Exception):
        raise_forever()


def test_retry_succeeds(monkeypatch):

    monkeypatch.setattr(time, 'sleep', mocksleep)

    params = {'times_to_fail': 2}
    @retry()
    def eventually_succeed(params):
        if params['times_to_fail'] > 0:
            params['times_to_fail'] -= 1
            raise
        else:
            return

    eventually_succeed(params)


def test_retry_limited_exceptions(monkeypatch):

    monkeypatch.setattr(time, 'sleep', mocksleep)

    params = {'times_to_fail': 2}

    @retry([BufferError])
    def eventually_succeed(params):
        if params['times_to_fail'] > 0:
            params['times_to_fail'] -= 1
            raise BufferError
        else:
            return

    eventually_succeed(params)


def test_retry_limited_exceptions_fail(monkeypatch):

    monkeypatch.setattr(time, 'sleep', mocksleep)

    params = {'times_to_fail': 2}

    @retry([LookupError])
    def fail(params):
        if params['times_to_fail'] > 0:
            params['times_to_fail'] -= 1
            raise BufferError
        else:
            return

    with pytest.raises(BufferError):
        fail(params)


def test_retry_attempts_insufficient(monkeypatch):

    monkeypatch.setattr(time, 'sleep', mocksleep)

    params = {'times_to_fail': 4}

    @retry([BufferError])
    def fail(params):
        if params['times_to_fail'] > 0:
            params['times_to_fail'] -= 1
            raise BufferError
        else:
            return

    with pytest.raises(BufferError):
        fail(params)


def test_retry_attempts_configurable(monkeypatch):

    monkeypatch.setattr(time, 'sleep', mocksleep)

    params = {'times_to_fail': 3}

    @retry([BufferError], retry_attempts=3)
    def eventually_succeeds(params):
        if params['times_to_fail'] > 0:
            params['times_to_fail'] -= 1
            raise BufferError
        else:
            return

    eventually_succeeds(params)



"""Configuration parameters for kvlayer.

These collectively implement the `yakonfig.Configurable` interface.

-----

This software is released under an MIT/X11 open source license.

Copyright 2012-2014 Diffeo, Inc.

"""
from __future__ import absolute_import

from kvlayer._client import STORAGE_CLIENTS
from yakonfig import ConfigurationError

config_name = 'kvlayer'
default_config = dict(
    storage_type = 'local',
    connection_pool_size = 2,
    max_consistency_delay = 120,
    replication_factor = 1,
    thrift_framed_transport_size_in_mb = 15,
    )

def add_arguments(parser):
    '''Add command line arguments to an argparse.ArgumentParser instance.'''
    parser.add_argument('--app-name',
                        help='name of app for namespace prefixing')
    parser.add_argument('--namespace',
                        help='namespace for prefixing table names')
    parser.add_argument('--storage-type',
                        help='backend type for kvlayer, e.g. "local" or "accumulo"')
    parser.add_argument('--storage-address', action='append',
                        dest='storage_addresses', metavar='HOST:PORT',
                        help='network addresses for kvlayer, can be repeated')
    parser.add_argument('--username', help='username for kvlayer accumulo')
    parser.add_argument('--password', help='password for kvlayer accumulo')

runtime_keys = dict(
    app_name = 'app_name',
    namespace = 'namespace',
    username = 'username',
    password = 'password',
    storage_type = 'storage_type',
    storage_addresses = 'storage_addresses',
    connection_pool_size = 'connection_pool_size',
    max_consistency_delay = 'max_consistency_delay',
    replication_factor = 'replication_factor',
    thrift_framed_transport_size_in_mb = 'thrift_framed_transport_size_in_mb',
    log_stats = 'log_stats',
    log_stats_interval_ops = 'log_stats_interval_ops',
    log_stats_interval_seconds = 'log_stats_interval_seconds',

    # these support testing and aren't exposed as command-line arguments
    kvlayer_filename = 'filename',
    kvlayer_copy_to_filename = 'copy_to_filename',
)

def check_config(config, name):
    if 'storage_type' not in config:
        raise ConfigurationError('{} must have a storage_type'
                                 .format(name))
    if config['storage_type'] not in STORAGE_CLIENTS:
        raise ConfigurationError('invalid {} storage_type {}'
                                 .format(name, config['storage_type']))
    if 'namespace' not in config:
        raise ConfigurationError('{} requires a namespace'.format(name))
    if 'app_name' not in config:
        raise ConfigurationError('{} requires an app_name'.format(name))



'''Mapping holding serializable values.

.. Your use of this software is governed by your license agreement.
   Copyright 2012-2014 Diffeo, Inc.

This is used to store dictionaries as values for kvlayer cells.

.. autoclass:: InstanceCollection
   :members:
   :special-members:
   :show-inheritance:

.. autofunction:: register

.. autoclass:: BlobCollection
   :members:
   :undoc-members:
   :show-inheritance:

.. autoclass:: Chunk
   :members:
   :show-inheritance:

'''
from __future__ import absolute_import
from abc import ABCMeta, abstractmethod
import collections
from cStringIO import StringIO
import importlib
import itertools

import streamcorpus
from kvlayer._exceptions import ProgrammerError, SerializationError
from kvlayer.instance_collection.ttypes import BlobCollection, TypedBlob

from thrift.transport import TTransport
from thrift.protocol.TBinaryProtocol import TBinaryProtocol, \
    TBinaryProtocolAccelerated
try:
    from thrift.protocol import fastbinary
    protocol = TBinaryProtocolAccelerated
except Exception, exc:
    protocol = TBinaryProtocol

class AugmentedStringIO(object):
    def __init__(self, blob):
        fh = StringIO(blob)
        self._fh = fh
        self.getvalue = fh.getvalue
        self.seek = fh.seek
        self.close = fh.close
        self.read = fh.read

    def readAll(self, sz):
        '''
        This method allows TBinaryProtocolAccelerated to actually function.

        Copied from here
        http://svn.apache.org/repos/asf/hive/trunk/service/lib/py/thrift/transport/TTransport.py
        '''
        buff = ''
        have = 0
        while (have < sz):
            chunk = self.read(sz - have)
            have += len(chunk)
            buff += chunk

            if len(chunk) == 0:
                raise EOFError()

        return buff

import json
import yaml
yaml.loads = yaml.load
yaml.dumps = yaml.dump

## global singleton dict of registered serializers
registered_serializers = dict(
    ## include "yaml" and "json" as defaults
    # must provide a callable that returns a thing with dumps/loads
    yaml=lambda: yaml,
    json=lambda: json,
    )

def register(name, serializer):
    global registered_serializers
    registered_serializers[name] = serializer

class InstanceCollection(collections.MutableMapping):
    '''A mapping of keys to serializable instance types.

    This maintains two sets of data: Python objects probably of some
    consistent type, and serialized versions of the same.  If this is
    created with only binary data, the Python objects will be created
    on demand.

    '''
    __metaclass__ = ABCMeta
    def __init__(self, data=None):
        '''Create a new InstanceCollection.

        `data` can provide data for the object at construction time.
        If missing or :const:`None`, the collection is initially
        empty.  If a :class:`BlobCollection`, the data in the
        collection is used for the initial values of the collection,
        and the keys in this collection are the keys from the blob.
        If a mapping, the collection is populated from the mapping,
        with the :meth:`__missing__` function used to generate
        serializations.  If a string, it contains the serialized
        Thrift version of a :class:`BlobCollection`, which is used as
        described above.

        :param data: initialization data

        '''
        self._instances = dict()
        if isinstance(data, BlobCollection):
            self.blobs = data
        elif isinstance(data, collections.Mapping):
            self._bc = BlobCollection()
            self._bc.collection_type = self._clsname
            self.update(data)
        else:
            self.loads(data)

    @staticmethod
    def from_data(data):
        '''Create a new InstanceCollection from data.

        `data` should be either a :class:`BlobCollection`, or a
        Thrift- serialized form of one.  The actual type constructed
        will be the :attr:`BlobCollection.collection_type` from the
        provided data.

        :param data: inbound `BlobCollection` or serialization thereof
        :return: new `InstanceCollection` or other serializable type

        '''
        if isinstance(data, BlobCollection):
            bc = data
        else:
            bc = BlobCollection()
            i_fh = AugmentedStringIO(data)
            i_transport = TTransport.TBufferedTransport(i_fh)
            i_protocol = protocol(i_transport)
            bc.read( i_protocol )

        target = bc.collection_type
        parts = target.split('.')
        modname = '.'.join(parts[:-1])
        cls = None
        try:
            mod = importlib.import_module(modname)
        except ValueError, e:
            mod = None
        except ImportError, e:
            mod = None
        if mod is not None:
            cls = getattr(mod, parts[-1], None)
        if cls is None:
            raise SerializationError("can't find collection class {!r}"
                                     .format(target))
        return cls(bc)

    @abstractmethod
    def __missing__(self, key):
        '''Provide a default value for some key.

        This should produce a default value and serialization for
        `key`, and insert it using :meth:`insert`.  Note that this is
        also called from e.g. :meth:`__setitem__()` when inserting a value
        for the first time.

        :param key: key of the mapping

        '''
        raise NotImplemented()

    def __repr__(self):
        return '{}(keys={!r}'.format(self.__class__.__name__,
                                     self._bc.typed_blobs.keys())

    @property
    def _clsname(self):
        '''Fully qualified module.Class name of this.'''
        return '{}.{}'.format(self.__class__.__module__,
                              self.__class__.__name__)

    @property
    def blobs(self):
        '''The :class:`BlobCollection` backing this :class:`InstanceCollection`.

        Any instances that have been fetched or added will be serialized
        into the :class:`BlobCollection` before returning it.

        '''
        for key, obj in self._instances.items():
            ## save the deserialized instance for repeated use
            serializer_name = self._bc.typed_blobs[key].serializer
            serializer_class = registered_serializers.get(serializer_name)
            if not serializer_class:
                raise SerializationError(
                    '{!r} has serializer={!r}, but that is not registered'
                    .format(key, serializer_name))
            serializer = serializer_class()
            if self._bc.typed_blobs[key].config:
                serializer.configure(self._bc.typed_blobs[key].config)
            self._bc.typed_blobs[key].blob = serializer.dumps(obj)
        return self._bc
    @blobs.setter
    def blobs(self, bc):
        '''Replace the :class:`BlobCollection` for this :class:`InstanceCollection`.

        All cached instances will be discarded.  The
        :attr:`BlobCollection.collection_type` of the collection must
        be the same as this object's runtime type.

        '''
        if bc.collection_type != self._clsname:
            raise SerializationError(
                'cannot create {} with a collection of {}'
                .format(self._clsname, bc.collection_type))
        self._bc = bc
        self._instances.clear()

    def loads(self, blob_collection_blob):
        '''Replaces the collection using a serialized :class:`BlobCollection`.

        All saved data in this object is discarded.  The provided
        string is parsed as a :class:`BlobCollection` and this collection is
        used as a source of serialized items.  If :const:`None` is provided
        then this collection is emptied.

        :param str blob_collection_blob: if not :const:`None`, then a serialized
          Thrift representation of a :class:`BlobCollection`

        '''
        bc = BlobCollection()
        if blob_collection_blob is None:
            bc.collection_type = self._clsname
        else:
            i_fh = AugmentedStringIO(blob_collection_blob)
            i_transport = TTransport.TBufferedTransport(i_fh)
            i_protocol = protocol(i_transport)
            bc.read( i_protocol )
        self.blobs = bc

    def dumps(self):
        '''Create a serialized :class:`BlobConnection` from this.

        All saved items in this object are serialized, and the
        resulting :class:`BlobCollection` is serialized and returned.
        Passing the resulting string to :meth:`loads` or to the class
        constructor should produce an equivalent object to `self`.

        :return: Thrift-serialized :class:`BlobConnection` string

        '''
        o_fh = StringIO()
        o_transport = TTransport.TBufferedTransport(o_fh)
        o_protocol = protocol(o_transport)
        self.blobs.write( o_protocol )
        o_transport.flush()
        return o_fh.getvalue()

    def __contains__(self, key):
        if key in self._instances:
            return True
        elif key in self._bc.typed_blobs:
            return True
        else:
            return False

    def __getitem__(self, key):
        if key not in self._instances:
            if key not in self._bc.typed_blobs:
                return self.__missing__(key)
            serializer_name = self._bc.typed_blobs[key].serializer
            serializer_class = registered_serializers.get(serializer_name)
            if not serializer_class:
                raise SerializationError(
                    '{!r} has serializer={!r}, but that is not registered'
                    .format(key, serializer_name))
            serializer = serializer_class()
            if self._bc.typed_blobs[key].config:
                serializer.configure(self._bc.typed_blobs[key].config)
            ## hold on to the deserialized instance for repeated use
            self._instances[key] = serializer.loads(self._bc.typed_blobs[key].blob)
        return self._instances[key]

    def __setitem__(self, key, value):
        self.__missing__(key)
        self._instances[key] = value

    def __delitem__(self, key):
        del self._instances[key]
        del self._bc.typed_blobs[key]

    def pop(self, key, default=None):
        if key in self:
            value = self.__getitem__(key)
            self._bc.typed_blobs.pop(key)
            self._instances.pop(key)
            return value
        elif default is not None:
            return default
        else:
            raise IndexError('%r is not in %r' % (key, self))

    def insert(self, key, value, serializer_name, config=None):
        '''Insert a value into the collection with an explicit serializer.

        This allows the caller to insert a value with the name of the
        serializer and an optional configuration block.  The
        serializer must be registered via the :func:`register`
        function.  In most cases, this will be called by the
        :meth:`__missing__` function, and ordinary :meth:`__setitem__`
        will correctly populate the item.

        :param key: mapping key
        :param value: mapping value
        :param str serializer_name: registered name of the binary
          serializer for `value`
        :param dict config: configuration associated with the object

        '''
        global registered_serializers
        if serializer_name not in registered_serializers:
            raise ProgrammerError('serializer_name=%r is not registered'
                                  % serializer_name)
        self._instances[key] = value
        if config is None:
            config = {}
        self._bc.typed_blobs[key] = TypedBlob(serializer = serializer_name,
                                        config = config)

    def __iter__(self):
        for key in self._bc.typed_blobs.keys():
            yield key

    def __len__(self):
        return len(self._bc.typed_blobs)


class Chunk(streamcorpus.Chunk):
    '''Chunk implementation that yields :class:`InstanceCollection` objects.

    The default `message` type is :class:`BlobCollection`, and you
    should generally use this.  :meth:`add()` adds, and
    :meth:`__iter__` produces, :class:`InstanceCollection` objects.

    In reality, any type that has a `blobs` property can be passed to
    :meth:`add()`; this property contains a :class:`BlobCollection`
    object.  When iterating, the :class:`BlobCollection` names a type
    in its :attr:`~BlobCollection.collection_type` field, and this
    must be a type that will accept the :class:`BlobCollection` as the
    only positional parameter in its constructor.

    '''
    def __init__(self, *args, **kwargs):
        kwargs.setdefault('message', BlobCollection)
        super(Chunk, self).__init__(*args, **kwargs)
    def add(self, msg):
        super(Chunk, self).add(msg.blobs)
    def __iter__(self):
        return itertools.imap(InstanceCollection.from_data,
                              super(Chunk, self).__iter__())



'''
Implementation of AbstractStorage using Postgres

Requires that you have the psycopg2 module installed in your environment:

  easy_install psycopg2
OR
  pip install psycopg2

This software is released under an MIT/X11 open source license.

Copyright 2012-2014 Diffeo, Inc.
'''
from __future__ import absolute_import
import logging
import re
import time

import psycopg2

from kvlayer._abstract_storage import AbstractStorage
from kvlayer._utils import split_key, make_start_key, make_end_key, join_key_fragments
from kvlayer._exceptions import ProgrammerError


logger = logging.getLogger(__name__)


# SQL strings in this module use python3 style string.format() formatting to substitute the table name into the command.


# this is not precisely right.
# http://www.postgresql.org/docs/9.3/static/sql-syntax-lexical.html#SQL-SYNTAX-IDENTIFIERS
# super-ascii 'letter' chars are also allowed.
# Perhaps in the all-unicode-all-the-time Python3 re package there will be
# better support for characters classes needed to specify this right.
_psql_identifier_re = re.compile(r'[a-z_][a-z0-9_$]*', re.IGNORECASE)

def _valid_namespace(x):
    return bool(_psql_identifier_re.match(x))


# kv_{namespace}
# table, key, value
_CREATE_TABLE = '''CREATE TABLE kv_{namespace} (
  t text,
  k bytea,
  v bytea,
  PRIMARY KEY (t, k)
);

CREATE FUNCTION upsert_{namespace}(tname TEXT, key BYTEA, data BYTEA) RETURNS VOID AS
$$
BEGIN
    LOOP
        -- first try to update the key
        UPDATE kv_{namespace} SET v = data WHERE t = tname AND k = key;
        IF found THEN
            RETURN;
        END IF;
        -- not there, so try to insert the key
        -- if someone else inserts the same key concurrently,
        -- we could get a unique-key failure
        BEGIN
            INSERT INTO kv_{namespace}(t,k,v) VALUES (tname, key, data);
            RETURN;
        EXCEPTION WHEN unique_violation THEN
            -- Do nothing, and loop to try the UPDATE again.
        END;
    END LOOP;
END;
$$
LANGUAGE plpgsql;
'''

_DROP_TABLE = "DROP FUNCTION upsert_{namespace}(TEXT,BYTEA,BYTEA)"
_DROP_TABLE_b = '''DROP TABLE kv_{namespace}'''

_CLEAR_TABLE = '''DELETE FROM kv_{namespace} WHERE t = %s'''

# use cursor.callproc() instead of SELECT query.
#_PUT = '''SELECT upsert_{namespace} (%s, %s, %s);'''

_GET = '''SELECT k, v FROM kv_{namespace} WHERE t = %s AND k = %s;'''

_GET_RANGE = '''SELECT k, v FROM kv_{namespace} WHERE t = %s AND k >= %s AND k <= %s ORDER BY k ASC LIMIT %s;'''
_GET_RANGE_FROM_START = '''SELECT k, v FROM kv_{namespace} WHERE t = %s AND k <= %s ORDER BY k ASC LIMIT %s;'''
_GET_RANGE_TO_END = '''SELECT k, v FROM kv_{namespace} WHERE t = %s AND k >= %s ORDER BY k ASC LIMIT %s;'''
_GET_ALL = '''SELECT k, v FROM kv_{namespace} WHERE t = %s ORDER BY k ASC LIMIT %s;'''

_DELETE = '''DELETE FROM kv_{namespace} WHERE t = %s AND k = %s;'''

# unused, we always _DELETE single records by key
#_DELETE_RANGE = '''DELETE FROM kv_{namespace} WHERE t = %s AND k >= %s AND K <= %s;'''

# TODO: use this query to list available namespaces
# select tablename from pg_catalog.pg_tables where tablename like 'kv_%';


MAX_BLOB_BYTES = 15000000


def _cursor_check_namespace_table(cursor, namespace):
    cursor.execute('SELECT 1 FROM pg_tables WHERE tablename ILIKE %s', ('kv_' + namespace,))
    return cursor.rowcount > 0


# detatch_on_exception
def detatch_on_exception(func):
    def _doe_fwrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except:
            try:
                args[0].close()
            except:
                logger.error('failed trying to close', exc_info=True)
                # don't re-raise the inner exception
            # re-raise the problem from func()
            raise
    return _doe_fwrapper


class PGStorage(AbstractStorage):
    def __init__(self):
        '''Initialize a storage instance for namespace.
        uses the single string specifier for a connectionn to a postgres db
http://www.postgresql.org/docs/current/static/libpq-connect.html#LIBPQ-PARAMKEYWORDS
        '''
        super(PGStorage, self).__init__()
        if not _valid_namespace(self._namespace):
            raise ProgrammerError('namespace must match re: %r' % (_psql_identifier_re.pattern,))
        self.storage_addresses = self._config['storage_addresses']
        if not self.storage_addresses:
            raise ProgrammerError('postgres kvlayer needs config["storage_addresses"]')
        self.connection = None
        self._scan_inner_limit = int(self._config.get('scan_inner_limit', 1000))

    def _conn(self):
        '''internal lazy connector'''
        if self.connection is None:
            self.connection = psycopg2.connect(self.storage_addresses[0])
            self.connection.autocommit = True
        return self.connection

    def _namespace_table_exists(self):
        conn = self._conn()
        with conn.cursor() as cursor:
            return _cursor_check_namespace_table(cursor, self._namespace)

    @detatch_on_exception
    def setup_namespace(self, table_names):
        '''creates tables in the namespace.  Can be run multiple times with
        different table_names in order to expand the set of tables in
        the namespace.

        :param table_names: Each string in table_names becomes the
        name of a table, and the value must be an integer specifying
        the number of UUIDs in the keys

        :type table_names: dict(str = int)
        '''
        self._table_names.update(table_names)
        self.normalize_namespaces(self._table_names)
        conn = self._conn()
        with conn.cursor() as cursor:
            if _cursor_check_namespace_table(cursor, self._namespace):
                # already exists
                logger.debug('namespace %r already exists, not creating', self._namespace)
                return
            cursor.execute(_CREATE_TABLE.format(namespace=self._namespace))

    @detatch_on_exception
    def delete_namespace(self):
        '''Deletes all data from namespace.'''
        conn = self._conn()
        with conn.cursor() as cursor:
            if not _cursor_check_namespace_table(cursor, self._namespace):
                logger.debug('namespace %r does not exist, not dropping', self._namespace)
                return
            try:
                cursor.execute(_DROP_TABLE.format(namespace=self._namespace))
                cursor.execute(_DROP_TABLE_b.format(namespace=self._namespace))
            except:
                logger.warn('error on delete_namespace(%r)', self._namespace, exc_info=True)

    @detatch_on_exception
    def clear_table(self, table_name):
        'Delete all data from one table'
        conn = self._conn()
        with conn.cursor() as cursor:
            cursor.execute(
                _CLEAR_TABLE.format(namespace=self._namespace),
                (table_name,)
            )

    @detatch_on_exception
    def put(self, table_name, *keys_and_values, **kwargs):
        '''Save values for keys in table_name.  Each key must be a
        tuple of UUIDs of the length specified for table_name in
        setup_namespace.

        :params batch_size: a DB-specific parameter that limits the
        number of (key, value) paris gathered into each batch for
        communication with DB.
        '''
        start_time = time.time()
        keys_size = 0
        values_size = 0
        num_keys = 0

        key_spec = self._table_names[table_name]
        conn = self._conn()
        with conn.cursor() as cursor:
            for kv in keys_and_values:
                ex = self.check_put_key_value(kv[0], kv[1], table_name, key_spec)
                if ex:
                    raise ex
                num_keys += 1
                keystr = join_key_fragments(kv[0], key_spec=key_spec)
                keys_size += len(keystr)
                values_size += len(kv[1])
                #logger.debug('put k=%r from %r', keystr, kv[0])
                keystr = psycopg2.Binary(keystr)
                cursor.callproc(
                    'upsert_{namespace}'.format(namespace=self._namespace),
                    (table_name, keystr, psycopg2.Binary(kv[1])))

        end_time = time.time()
        num_values = num_keys

        self.log_put(table_name, start_time, end_time, num_keys, keys_size, num_values, values_size)

    @detatch_on_exception
    def get(self, table_name, *keys, **kwargs):
        '''Yield tuples of (key, value) from querying table_name for
        items with specified keys.
        '''
        start_time = time.time()
        num_keys = 0
        keys_size = 0
        num_values = 0
        values_size = 0

        key_spec = self._table_names[table_name]
        cmd = _GET.format(namespace=self._namespace)
        conn = self._conn()
        with conn.cursor() as cursor:
            for key in keys:
                num_keys += 1
                bkey = join_key_fragments(key, key_spec=key_spec)
                keys_size += len(bkey)
                bkey = psycopg2.Binary(bkey)
                cursor.execute(cmd, (table_name, bkey))
                if not (cursor.rowcount > 0):
                    yield key, None
                    continue
                results = cursor.fetchmany()
                while results:
                    for row in results:
                        val = row[1]
                        if isinstance(val, buffer):
                            if len(val) > MAX_BLOB_BYTES:
                                logger.error('key=%r has blob of size %r over limit of %r', row[0], len(val), MAX_BLOB_BYTES)
                                continue  # TODO: raise instead of drop?
                            val = val[:]
                        keyraw = row[0]
                        if isinstance(keyraw, buffer):
                            keyraw = keyraw[:]
                        num_values += 1
                        values_size += len(val)
                        yield split_key(keyraw, key_spec), val
                    results = cursor.fetchmany()

        end_time = time.time()
        self.log_get(table_name, start_time, end_time, num_keys, keys_size, num_values, values_size)

    @detatch_on_exception
    def scan(self, table_name, *key_ranges, **kwargs):
        '''Yield tuples of (key, value) from querying table_name for
        items with keys within the specified ranges.  If no key_ranges
        are provided, then yield all (key, value) pairs in table.

        :type key_ranges: (((UUID, ...), (UUID, ...)), ...)
                            ^^^^^^^^^^^^^^^^^^^^^^^^
                            start        finish of one range
        '''
        start_time = time.time()
        num_keys = 0
        keys_size = 0
        values_size = 0

        key_spec = self._table_names[table_name]
        if not key_ranges:
            key_ranges = [['', '']]
        def _pgkeyrange(kr):
            return (table_name, kmin, kmax)
        conn = self._conn()
        with conn.cursor() as cursor:
            for kmin, kmax in key_ranges:
                for keyraw, rval in self._scan_subscan_kminmax(key_spec, cursor, table_name, kmin, kmax):
                    rkey = split_key(keyraw, key_spec)
                    yield rkey, rval

                    num_keys += 1
                    keys_size += len(keyraw)
                    values_size += len(rval)

        end_time = time.time()
        num_values = num_keys
        self.log_scan(table_name, start_time, end_time, num_keys, keys_size, num_values, values_size)

    def _scan_subscan_kminmax(self, key_spec, cursor, table_name, kmin, kmax):
        prevkey = None
        while True:
            count = 0
            for keyraw, rval in self._scan_kminmax(key_spec, cursor, table_name, kmin, kmax):
                count += 1
                if keyraw != prevkey:
                    # don't double-return an edge value
                    yield keyraw, rval
                prevkey = keyraw
            if count < self._scan_inner_limit:
                # we didn't get up to limit, we must be done
                return
            # else, we hit limit, we need to scan for more
            kmin = split_key(prevkey, key_spec)

    def _scan_kminmax(self, key_spec, cursor, table_name, kmin, kmax):
        if kmin:
            start = make_start_key(kmin, key_spec=key_spec)
            start = psycopg2.Binary(start)
        else:
            start = None
        if kmax:
            finish = make_end_key(kmax, key_spec=key_spec)
            finish = psycopg2.Binary(finish)
        else:
            finish = None
        #logger.debug('pg t=%r %r<=k<=%r (%s<=k<=%s)', table_name, kmin, kmax, start, finish)
        if start:
            if finish:
                cmd = _GET_RANGE.format(namespace=self._namespace)
                cursor.execute(cmd, (table_name, start, finish, self._scan_inner_limit))
            else:
                cmd = _GET_RANGE_TO_END.format(namespace=self._namespace)
                cursor.execute(cmd, (table_name, start, self._scan_inner_limit))
        else:
            if finish:
                cmd = _GET_RANGE_FROM_START.format(namespace=self._namespace)
                cursor.execute(cmd, (table_name, finish, self._scan_inner_limit))
            else:
                cmd = _GET_ALL.format(namespace=self._namespace)
                cursor.execute(cmd, (table_name, self._scan_inner_limit))
        #logger.debug('%r rows from %r', cursor.rowcount, cmd)
        if not (cursor.rowcount > 0):
            return
        results = cursor.fetchmany()
        while results:
            for row in results:
                val = row[1]
                if isinstance(val, buffer):
                    if len(val) > MAX_BLOB_BYTES:
                        logger.error('key=%r has blob of size %r over limit of %r', row[0], len(val), MAX_BLOB_BYTES)
                        continue  # TODO: raise instead of drop?
                    val = val[:]
                keyraw = row[0]
                if isinstance(keyraw, buffer):
                    keyraw = keyraw[:]
                yield keyraw, val
            results = cursor.fetchmany()

    @detatch_on_exception
    def delete(self, table_name, *keys, **kwargs):
        '''Delete all (key, value) pairs with specififed keys

        :params batch_size: a DB-specific parameter that limits the
        number of (key, value) paris gathered into each batch for
        communication with DB.
        '''
        start_time = time.time()
        num_keys = 0
        keys_size = 0

        key_spec = self._table_names[table_name]
        delete_statement_args = []
        for k in keys:
            if len(k) != len(key_spec):
                raise Exception('invalid key has %s uuids but wanted %s: %r' % (len(k), len(key_spec), k))
            joined_key = join_key_fragments(k, key_spec=key_spec)
            num_keys += 1
            keys_size += len(joined_key)
            delete_statement_args.append(
                (table_name, psycopg2.Binary(joined_key))
            )

        conn = self._conn()
        with conn.cursor() as cursor:
            cursor.executemany(
                _DELETE.format(namespace=self._namespace),
                delete_statement_args)

        end_time = time.time()
        self.log_delete(table_name, start_time, end_time, num_keys, keys_size)

    # don't mark this one detatch_on_exception, that would be silly
    def close(self):
        '''
        close connections and end use of this storage client
        '''
        if self.connection:
            try:
                self.connection.close()
            finally:
                self.connection = None


# run this to cleanup any cruft from kvlayer unit tests
CLEAN_TESTS = '''
CREATE OR REPLACE FUNCTION clean_tests() RETURNS VOID AS
$$
DECLARE
  argtypes text;
  tat text;
  toid oid;
  pnargs pg_proc%ROWTYPE;
  pt pg_tables%ROWTYPE;
BEGIN
  FOR pnargs IN SELECT * from pg_proc where proname like '%upsert_test_%' LOOP
    SELECT typname FROM pg_type WHERE oid = pnargs.proargtypes[0] INTO argtypes;
    FOR i in 1..array_upper(pnargs.proargtypes,1) LOOP
      SELECT typname FROM pg_type WHERE oid = pnargs.proargtypes[i] INTO tat;
      argtypes := argtypes || ',' || tat;
    END LOOP;
    EXECUTE 'DROP FUNCTION ' || pnargs.proname || '(' || argtypes || ');';
  END LOOP;
  FOR pt IN SELECT * FROM pg_tables WHERE tablename LIKE '%kv_test_%' LOOP
    EXECUTE 'DROP TABLE ' || pt.tablename || ';';
  END LOOP;
END
$$ LANGUAGE plpgsql;


SELECT clean_tests();
'''



from __future__ import absolute_import
import uuid

import pytest

import kvlayer
from kvlayer._local_memory import LocalStorage
import yakonfig

@pytest.yield_fixture(scope='module')
def local_storage():
    config_yaml = """
kvlayer:
    storage_type: local
    storage_addresses: []
    namespace: test
    app_name: test
"""
    with yakonfig.defaulted_config([kvlayer], yaml=config_yaml):
        local_storage = LocalStorage()
        yield local_storage

def test_local_storage_singleton(local_storage):
    local_storage.setup_namespace(dict(meta=1))
    keys_and_values = ((uuid.uuid4(),), b'hi')
    local_storage.put('meta', keys_and_values)
    key_range = (keys_and_values[0], keys_and_values[0])
    meta = list(local_storage.scan('meta', key_range))
    assert meta[0][1] == b'hi'
    local_storage2 = LocalStorage()
    local_storage2.setup_namespace(dict(meta=1))
    meta = list(local_storage2.scan('meta', key_range))
    assert meta[0][1] == b'hi'

def test_get(local_storage):
    local_storage.setup_namespace(dict(meta=1))
    value = b'hi2'
    keys = tuple([uuid.uuid4()])
    local_storage.put('meta', tuple((keys, value)))
    res = list(local_storage.get('meta', keys))
    for res_key, res_value in res:
        if res_key == keys:
            assert res_value == value
    assert keys in [rk for rk, rv in res]

def test_get_complex_keys(local_storage):
    local_storage.setup_namespace(dict(meta=2))
    value = b'hi2'
    key = (uuid.uuid4(), uuid.uuid4())
    local_storage.put('meta', tuple((key, value)))
    res = list(local_storage.get('meta', key))
    for res_key, res_value in res:
        if res_key == key:
            assert res_value == value
    assert key in [rk for rk, rv in res]

def test_delete_namespace(local_storage):
    """Test that delete_namespace() actually clears the shared storage"""
    u = (uuid.uuid4(),)

    local_storage.setup_namespace(dict(meta=1))
    local_storage.put('meta', (u, b'hi'))
    assert list(local_storage.get('meta', u)) == [(u, b'hi')]
    local_storage.delete_namespace()

    local_storage2 = LocalStorage()
    local_storage2.setup_namespace(dict(meta=1))
    assert list(local_storage2.get('meta', u)) == [(u, None)]



"""Basic functional tests for all of the kvlayer backends.

Your use of this software is governed by your license agreement.

Copyright 2012-2014 Diffeo, Inc.

"""
from __future__ import absolute_import
import errno
import logging
import os
import pdb
import random
import re
import cStringIO as StringIO
import sys
import time
import uuid

import py
import pytest
import yaml

from pytest_diffeo import redis_address
import kvlayer
from kvlayer import BadKey
import yakonfig

logger = logging.getLogger(__name__)

# each backend needs to be fitted with instrumatation.
# TOOD: add to redis and accumulo
#backends = ['local', 'filestorage', 'redis', 'accumulo']
backends = ['local', 'filestorage']
try:
    import psycopg2
    backends.append('postgres')
except ImportError, e:
    # backends.dont_append('postgres')
    pass

@pytest.fixture(scope='module',
                params=backends)
def backend(request):
    return request.param

@pytest.yield_fixture(scope='function')
def client(backend, request, tmpdir, namespace_string):
    config_path = str(request.fspath.dirpath('config_{}.yaml'.format(backend)))
    statsfile = StringIO.StringIO()
    params = dict(
        app_name='kvlayer',
        namespace=namespace_string,
        log_stats=statsfile,
        log_stats_interval_ops=1,
        blagh='hoo haa',
    )

    # this is hacky but must go somewhere
    if backend == 'filestorage':
        local = tmpdir.join('local')
        with local.open('w') as f: pass
        params['kvlayer_filename'] = str(local)

    if backend == 'redis':
        params['storage_addresses'] = [ redis_address(request) ]

    with yakonfig.defaulted_config([kvlayer], filename=config_path,
                                   params=params):
        client = kvlayer.client()
        client.delete_namespace()
        yield client
        client.delete_namespace()


# Deeply invasive test that knows lots of internal structure.
def test_stats(client):
    client.setup_namespace(dict(t1=(str,str), t2=(str,str,str)))
    assert 0 == len(list(client.scan('t1')))

    assert client._log_stats.scan.num_ops == 1
    assert client._log_stats.scan.by_table['t1'].num_ops == 1

    u1, u2, u3 = 'u1', 'u2', 'u3'
    client.put('t1', ((u1, u2), b'88'))
    client.put('t2', ((u1, u2, u3), b'88'))

    assert client._log_stats.put.num_ops == 2
    assert client._log_stats.put.by_table['t1'].num_ops == 1
    assert client._log_stats.put.by_table['t2'].num_ops == 1

    assert 1 == len(list(client.scan('t1')))
    assert 1 == len(list(client.scan('t1', ((u1,), (u1,)))))

    client.delete('t1', (u1, u2))

    assert client._log_stats.delete.num_ops == 1
    assert client._log_stats.delete.by_table['t1'].num_ops == 1
    assert client._log_stats.delete.by_table.get('t2') is None

    assert 0 == len(list(client.scan('t1')))
    assert 0 == len(list(client.scan('t1', ((u1,), (u1,)))))
    assert 0 == len(list(client.scan('t2', ((u2,), (u3,)))))

    assert client._log_stats.scan.num_ops == 6
    assert client._log_stats.scan.by_table['t1'].num_ops == 5
    assert client._log_stats.scan.by_table['t2'].num_ops == 1

    client._log_stats.flush()
    f = client._log_stats._f
    raw = f.getvalue()
    assert len(re.findall('put:', raw)) == 10
    logger.info('stats: %s', raw)
    client._log_stats.close()
    client._log_stats = None




'''
applications using kvlayer can call kvlayer.add_arguments(parser) to
get sensible defaults and help messages for command line options.

This software is released under an MIT/X11 open source license.

Copyright 2012-2014 Diffeo, Inc.
'''
from cStringIO import StringIO
import logging
import uuid

import pytest

import kvlayer
from kvlayer._client import Actions
import yakonfig

logger = logging.getLogger(__name__)

@pytest.yield_fixture
def actions(namespace_string):
    with yakonfig.defaulted_config([kvlayer], params={
            'app_name': 'diffeo',
            'namespace': namespace_string,
            'storage_type': 'local',
            'storage_addresses': [],
    }):
        a = Actions(stdout=StringIO())
        yield a
        a.client.delete_namespace()

@pytest.fixture
def a_key(actions):
    client = actions.client
    client.setup_namespace(dict(t1=1))
    k1 = (uuid.uuid4(),)
    client.put('t1', (k1, 'some data'))
    return k1

def test_delete(actions, a_key):
    assert list(actions.client.get('t1', a_key)) == [(a_key, 'some data')]
    actions.runcmd('delete', ['-y'])
    assert actions.stdout.getvalue().startswith("deleting namespace '")
    assert list(actions.client.get('t1', a_key)) == [(a_key, None)]

def test_keys(actions, a_key):
    actions.runcmd('keys', ['t1', '1'])
    assert actions.stdout.getvalue() == repr(a_key) + '\n'

def test_keys_none(actions):
    actions.runcmd('keys', ['t1', '1'])
    assert actions.stdout.getvalue() == ''

def test_keys_by_uuid(actions, a_key):
    actions.runcmd('keys', ['t1', 'uuid'])
    assert actions.stdout.getvalue() == repr(a_key) + '\n'

def test_get(actions, a_key):
    actions.runcmd('get', ['t1', '1', a_key[0].hex])
    assert actions.stdout.getvalue() == "'some data'\n"

def test_get_none(actions, a_key):
    b_key = uuid.UUID(int=a_key[0].int + 1)
    actions.runcmd('get', ['t1', '1', b_key.hex])
    assert (actions.stdout.getvalue() ==
            'No values for key (' + repr(b_key) + ',).\n')



'''
This software is released under an MIT/X11 open source license.

Copyright 2012-2014 Diffeo, Inc.
'''
from __future__ import absolute_import
import argparse
from cStringIO import StringIO
import logging
import sys
import termios
import tty
import uuid

import yaml

import kvlayer
from kvlayer._accumulo import AStorage
from kvlayer._cassandra import CStorage
from kvlayer._local_memory import LocalStorage
from kvlayer._file_storage import FileStorage
from kvlayer._redis import RedisStorage
import yakonfig
from yakonfig.cmd import ArgParseCmd

try:
    from kvlayer._postgres import PGStorage
except ImportError:
    PGStorage = None

logger = logging.getLogger(__name__)

STORAGE_CLIENTS = dict(
    ## these strings deinfe the external API for selecting the kvlayer
    ## storage backends
    cassandra=CStorage,
    accumulo=AStorage,
    local=LocalStorage,
    filestorage=FileStorage,
    redis=RedisStorage
)

if PGStorage:
    STORAGE_CLIENTS['postgres'] = PGStorage

def client():
    '''
    constructs a storage client for the storage_type specified in config
    '''
    config = yakonfig.get_global_config('kvlayer')
    try:
        return STORAGE_CLIENTS[config['storage_type']]()
    except Exception, exc:
        logger.critical('config = %r' % config, exc_info=True)
        raise

class Actions(ArgParseCmd):
    def __init__(self, *args, **kwargs):
        ArgParseCmd.__init__(self, *args, **kwargs)
        self.prompt = 'kvlayer> '
        self._client = None

    @property
    def client(self):
        if self._client is None:
            self._client = kvlayer.client()
        return self._client

    def args_delete(self, parser):
        parser.add_argument('-y', '--yes', default=False, action='store_true',
                            dest='assume_yes',
                            help='assume "yes" and require no input for '
                            'confirmation questions.')
    def do_delete(self, args):
        '''delete all tables in the current namespace'''
        namespace = self.client._namespace
        if not args.assume_yes:
            response = raw_input('Delete everything in {!r}?  Enter namespace: '
                                 .format(namespace))
            if response != namespace:
                self.stdout.write('not deleting anything\n')
                return
        self.stdout.write('deleting namespace {!r}\n'.format(namespace))
        self.client.delete_namespace()

    @staticmethod
    def _schema(s):
        n = { 'uuid': uuid.UUID, 'int': int, 'long': long, 'str': str }
        if s.isdigit(): return (uuid.UUID,) * int(s)
        if s in n: return (n[s],)
        if s.startswith('(') and s.endswith(')'): s=s[1:-1]
        parts = s.split(',')
        return tuple(n[p] for p in parts)

    def args_keys(self, parser):
        parser.add_argument('table', help='name of kvlayer table')
        parser.add_argument('schema', help='description of table keys',
                            type=self._schema)
    def do_keys(self, args):
        '''list all keys in a single table'''
        self.client.setup_namespace({ args.table: args.schema })
        for k,v in self.client.scan(args.table):
            self.stdout.write('{!r}\n'.format(k))

    def args_get(self, parser):
        parser.add_argument('table', help='name of kvlayer table')
        parser.add_argument('schema', help='description of table keys',
                            type=self._schema)
        parser.add_argument('keys', nargs='+',
                            help='key to fetch')
    def do_get(self, args):
        '''get values from a single key'''
        self.client.setup_namespace({ args.table: args.schema })
        key = tuple(f(x) for f, x in zip(args.schema, args.keys))
        for k,v in self.client.get(args.table, key):
            if v is None:
                self.stdout.write('No values for key {!r}.\n'.format(k))
            else:
                self.stdout.write('{!r}\n'.format(v))

def main():
    parser = argparse.ArgumentParser()
    action = Actions()
    action.add_arguments(parser)
    args = yakonfig.parse_args(parser, [yakonfig, kvlayer])
    action.main(args)

if __name__ == '__main__':
    main()



#!/usr/bin/env python

import os
import sys
import fnmatch
import subprocess

## prepare to run PyTest as a command
from distutils.core import Command
from distutils.dir_util import remove_tree

from setuptools import setup, find_packages

from version import get_git_version
VERSION, SOURCE_LABEL = get_git_version()
PROJECT = 'kvlayer'
URL = 'http://diffeo.com'
AUTHOR = 'Diffeo, Inc.'
AUTHOR_EMAIL = 'support@diffeo.com'
DESC = 'table-oriented abstraction layer over key-value stores'

def read_file(file_name):
    file_path = os.path.join(
        os.path.dirname(__file__),
        file_name
        )
    return open(file_path).read()

def recursive_glob(treeroot, pattern):
    results = []
    for base, dirs, files in os.walk(treeroot):
        goodfiles = fnmatch.filter(files, pattern)
        results.extend(os.path.join(base, f) for f in goodfiles)
    return results

def recursive_glob_with_tree(treeroot, pattern):
    results = []
    for base, dirs, files in os.walk(treeroot):
        goodfiles = fnmatch.filter(files, pattern)
        one_dir_results = []
        for f in goodfiles:
            one_dir_results.append(os.path.join(base, f))
        results.append((base, one_dir_results))
    return results

class Thrift(Command):
    '''run thrift'''
    description = 'run thrift generator from IDL to generated python'

    user_options = [
        ('force', 'f',
         "run all the build commands even if we don't need to")
        ]

    boolean_options = ['force']

    def initialize_options(self):
        self.force = 0
    def finalize_options(self):
        pass
    def run(self):
        self.maybe_thrift_gen('kvlayer/instance_collection/blob_collection.thrift', 'kvlayer/instance_collection')

    def maybe_thrift_gen(self, thrift_src, outdir, renamefunc=None):
        if renamefunc is None:
            renamefunc = lambda x: x
        self.make_file(
            thrift_src,
            os.path.join(outdir, renamefunc('ttypes.py')),
            self._run_thrift,
            [thrift_src, outdir, renamefunc])

    def _run_thrift(self, thrift_src, outdir, renamefunc):
        self.spawn(['thrift', '--gen', 'py:new_style,slots', thrift_src])
        for fname in ('constants.py', 'ttypes.py'):
            self.copy_file('gen-py/kvlayer/instance_collection/blob_collection/' + fname, os.path.join(outdir, renamefunc(fname)))
        remove_tree('gen-py')

def _myinstall(pkgspec):
    subprocess.check_call(['pip', 'install', pkgspec])

class PyTest(Command):
    '''run py.test'''

    description = 'runs py.test to execute all tests'

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        if self.distribution.install_requires:
            for ir in self.distribution.install_requires:
                _myinstall(ir)
        if self.distribution.tests_require:
            for ir in self.distribution.tests_require:
                _myinstall(ir)

        errno = subprocess.call(['py.test', '-n', '3', '-s', 'kvlayer/tests', '--runslow', '--runperf'])
        raise SystemExit(errno)

setup(
    name=PROJECT,
    version=VERSION,
    description=DESC,
    long_description=read_file('README.md'),
    author=AUTHOR,
    license='MIT/X11 license http://opensource.org/licenses/MIT',
    author_email=AUTHOR_EMAIL,
    url=URL,
    packages=find_packages(),
    cmdclass={
        'test': PyTest,
        'thrift': Thrift,
    },
    # We can select proper classifiers later
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Topic :: Utilities',
        'License :: OSI Approved :: MIT License',  ## MIT/X11 license http://opensource.org/licenses/MIT
    ],
    # psycopg2 may be commented out to operate without postgres support
    # Might be nice to have a conditional install_requires which determines
    # what underlying database connectors can actually be supported on the install
    # box. Perhaps something like:
    # http://stackoverflow.com/questions/14036181/provide-a-complex-condition-in-install-requires-python-setuptoolss-setup-py
    install_requires=[
        'yakonfig >= 0.5.0',
        #'psycopg2',
        'pycassa >= 1.10',
        'pyaccumulo >= 1.5.0.5',
        'pyyaml',
        'cql',
        'streamcorpus >= 0.3.23',
        'redis',
        # test dependencies
        'guppy',
        'pytest',
        'pytest-cov',
        'pytest-xdist',
        'pytest-timeout',
        'pytest-incremental',
        'pytest-capturelog',
        'pytest-diffeo >= 0.1.4'
    ],
    entry_points={
        'console_scripts': [
            'kvlayer = kvlayer._client:main',
            'kvlayer_test = kvlayer.tests.run:main',
            'kvlayer_throughput_test = kvlayer.tests.test_throughput:main',
        ]
    },
    include_package_data = True,
)



'''
Implementation of AbstractStorage using Cassandra

Your use of this software is governed by your license agreement.

Copyright 2012-2013 Diffeo, Inc.
'''

import re
import logging
from kvlayer._decorators import retry
from kvlayer._exceptions import ProgrammerError
from kvlayer._abstract_storage import AbstractStorage
from pyaccumulo import Accumulo, Mutation, Range, BatchWriter
from pyaccumulo.iterators import RowDeletingIterator
from pyaccumulo.proxy.ttypes import IteratorScope, AccumuloSecurityException
from _utils import split_key, make_start_key, make_end_key, join_key_fragments

logger = logging.getLogger('kvlayer')


class AStorage(AbstractStorage):
    '''
    Accumulo storage implements kvlayer's AbstractStorage, which
    manages a set of tables as specified to setup_namespace
    '''
    def __init__(self):
        super(AStorage, self).__init__()

        self._connected = False
        addresses = self._config.get('storage_addresses', [])
        if not addresses:
            raise ProgrammerError('config lacks storage_addresses')

        logger.info('accumulo thrift proxy supports only one address, using '
                    'first: %r' % addresses)
        address = addresses[0]
        if ':' not in address:
            self._host = address
            self._port = 50096
        else:
            self._host, self._port = address.split(':')
            self._port = int(self._port)

        ## The following are all parameters to the accumulo
        ## batch interfaces.
        self._max_memory = self._config.get('accumulo_max_memory', 1000000)
        self._timeout_ms = self._config.get('accumulo_timeout_ms', 30000)
        self._threads = self._config.get('accumulo_threads', 10)
        self._latency_ms = self._config.get('accumulo_latency_ms', 10)

        ## Acumulo storage requires a username and password
        self._user = self._config.get('username', None)
        self._password = self._config.get('password', None)
        if not self._user and self._password:
            raise ProgrammerError('accumulo storage requires '
                                  'username/password')

        self.thrift_framed_transport_size_in_mb = \
            self._config['thrift_framed_transport_size_in_mb']

        self._conn = None

    @property
    def conn(self):
        if not self._conn:
            logger.info('connecting to Accumulo')
            self._conn = Accumulo(self._host, self._port,
                                  self._user, self._password)
            self._connected = True
        return self._conn

    def _ns(self, table):
        '''
        accumulo does not have "namespaces" as a concept per se, so we
        achieve the same effect by creating tables with the namespace
        string appended to the end of the table name
        '''
        return '%s_%s_%s' % (self._app_name, self._namespace, table)

    def _create_table(self, table):
        logger.info('creating %s', self._ns(table))

        self.conn.create_table(self._ns(table))
        logger.debug('conn.created_table(%s)', self._ns(table))
        self.conn.client.setTableProperty(self.conn.login,
                                          self._ns(table),
                                          'table.bloom.enabled',
                                          'true')
        logger.debug("conn.client.setTableProperty(%r, %s, table.bloom.enabled', 'true'",
                     self.conn.login, self._ns(table))

        i = RowDeletingIterator()
        scopes = set([IteratorScope.SCAN, IteratorScope.MINC,
                      IteratorScope.MAJC])
        i.attach(self.conn, self._ns(table), scopes)
        logger.debug('i.attach(%r, %s, %r)', self.conn, self._ns(table), scopes)

    def setup_namespace(self, table_names):
        '''creates tables in the namespace.  Can be run multiple times with
        different table_names in order to expand the set of tables in
        the namespace.
        '''
        logger.debug('creating tables: %r', table_names)
        self._table_names.update(table_names)
        self.normalize_namespaces(self._table_names)
        for table in table_names:
            if not self.conn.table_exists(self._ns(table)):
                self._create_table(table)

    def delete_namespace(self):
        '''
        delete all of the tables within namespace
        '''
        logger.debug('getting list of tables')
        tables = self.conn.list_tables()
        logger.debug('finding tables to delete from %s: %r',
                     self._ns(''), tables)
        tables_to_delete = [x for x in tables if re.search(self._ns(''), x)]
        for table in tables_to_delete:
            self.conn.delete_table(table)

    def clear_table(self, table_name):
        self.conn.delete_table(self._ns(table_name))
        self.conn.create_table(self._ns(table_name))

    @retry([AccumuloSecurityException])
    def put(self, table_name, *keys_and_values, **kwargs):
        key_spec = self._table_names[table_name]
        cur_bytes = 0
        batch_writer = BatchWriter(conn=self.conn,
                                   table=self._ns(table_name),
                                   max_memory=self._max_memory,
                                   latency_ms=self._latency_ms,
                                   timeout_ms=self._timeout_ms,
                                   threads=self._threads)
        for key, blob in keys_and_values:
            ex = self.check_put_key_value(key, blob, table_name, key_spec)
            if ex:
                raise ex
            if (len(blob) + cur_bytes >=
                    self.thrift_framed_transport_size_in_mb * 2 ** 19):
                logger.debug('len(blob)=%d + cur_bytes=%d >= '
                             'thrift_framed_transport_size_in_mb/2 = %d',
                             len(blob), cur_bytes,
                             self.thrift_framed_transport_size_in_mb * 2 ** 19)
                logger.debug('pre-emptively sending only what has been '
                             'batched, and will send this item in next batch.')
                batch_writer.flush()
                cur_bytes = 0
            mut = Mutation(join_key_fragments(key, key_spec=key_spec))
            mut.put(cf='', cq='', val=blob)
            batch_writer.add_mutation(mut)
            cur_bytes += len(blob)
        batch_writer.close()

    def scan(self, table_name, *key_ranges, **kwargs):
        key_spec = self._table_names[table_name]
        if not key_ranges:
            key_ranges = [['', '']]
        for start_key, stop_key in key_ranges:
            total_count = 0
            specific_key_range = bool(start_key or stop_key)
            if specific_key_range:
                # Accumulo treats None as a negative-infinity or
                # positive-infinity key as needed for starts and ends
                # of ranges.
                #
                # pyaccumulo has a bug at present 20140228_171555
                # which causes it to never do a '>=' scan, so we must
                # decrement the start key to include the start key
                # which necessary for a get() operation.
                # https://github.com/accumulo/pyaccumulo/issues/14
                if not start_key:
                    srow = None
                else:
                    srow = make_start_key(start_key, key_spec=key_spec)
                    srow = _string_decrement(srow)
                if not stop_key:
                    erow = None
                else:
                    erow = make_end_key(stop_key, key_spec=key_spec)
                key_range = Range(srow=srow, erow=erow, sinclude=True, einclude=True)
                scanner = self.conn.scan(self._ns(table_name),
                                         scanrange=key_range)
            else:
                scanner = self.conn.scan(self._ns(table_name))

            for row in scanner:
                total_count += 1
                yield split_key(row.row, key_spec), row.val

    def get(self, table_name, *keys, **kwargs):
        for key in keys:
            gen = self.scan(table_name, (key, key))
            v = None
            for kk, vv in gen:
                if kk == key:
                    v = vv
            yield key, v

    def close(self):
        self._connected = False
        if hasattr(self, 'pool') and self.pool:
            self.pool.dispose()

    def __del__(self):
        self.close()

    def delete(self, table_name, *keys, **kwargs):
        batch_writer = BatchWriter(conn=self.conn,
                                   table=self._ns(table_name),
                                   max_memory=self._max_memory,
                                   latency_ms=self._latency_ms,
                                   timeout_ms=self._timeout_ms,
                                   threads=self._threads)
        for key in keys:
            mut = Mutation(join_key_fragments(key, key_spec=self._table_names[table_name]))
            mut.put(cf='', cq='', val='DEL_ROW')
            batch_writer.add_mutation(mut)
        batch_writer.close()


def _string_decrement(x):
    if len(x) < 1:
        return None  # None is before all keys, aka negative infinity
    pre = x[:-1]
    post = ord(x[-1])
    if post > 0:
        return pre + chr(post - 1) + '\xff'
    else:
        pre = _string_decrement(pre)
        if pre is None:
            return None
        return pre + '\xff'



import logging

import pytest

import kvlayer
import yakonfig

logger = logging.getLogger(__name__)

@pytest.fixture(scope='function')
def namespace(request, namespace_string):
    def fin():
        logger.info('tearing down %s...', namespace_string)
        try:
            config = yakonfig.get_global_config('kvlayer')
            ## this is probably already in the config
            config['namespace'] = namespace_string
            client = kvlayer.client(config)
            client.delete_namespace()
            logger.info('finished tearing down %s.', namespace_string)
        except KeyError:
            logger.warn('%s not configured in this process; cannot guess config', namespace_string)
        except Exception, exc:
            logger.error('failed to tear down %s', namespace_string, exc_info=True)
    request.addfinalizer(fin)
    return namespace_string
